# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import re
from copy import deepcopy
from dataclasses import dataclass
from itertools import groupby, combinations
from typing import List, Dict, Optional, Tuple

import numpy as np
from pydefect.analyzer.band_edge_states import BandEdgeStates
from pydefect.analyzer.calc_results import CalcResults
from pydefect.corrections.abstract_correction import Correction
from pydefect.corrections.manual_correction import NoCorrection
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import Element, IStructure
from scipy.spatial import HalfspaceIntersection
import dash_html_components as html


@dataclass
class SingleDefectEnergy:
    name: str
    charge: int
    energy: float
    correction: float

    @property
    def corrected_energy(self):
        return self.energy + self.correction


@dataclass
class DefectEnergy:
    name: str
    charges: List[int]
    energies: List[float]
    corrections: List[float]

    def stable_charges(self, ef_min, ef_max):
        return set(self.cross_points(ef_min, ef_max).charges)

    @property
    def corrected_energies(self):
        result = []
        for energy, correction in zip(self.energies, self.corrections):
            result.append(energy + correction)
        return result

    def cross_points(self, ef_min, ef_max, base_e=0.0):
        large_minus_number = -1e4
        half_spaces = []
        for charge, corr_energy in zip(self.charges, self.corrected_energies):
            half_spaces.append([-charge, 1, -corr_energy])

        half_spaces.append([-1, 0, ef_min])
        half_spaces.append([1, 0, -ef_max])
        half_spaces.append([0, -1, large_minus_number])

        feasible_point = np.array([(ef_min + ef_max) / 2, -1e3])

        hs = HalfspaceIntersection(np.array(half_spaces), feasible_point)
        boundary_points = []
        inner_cross_points = []
        for intersection in hs.intersections:
            x, y = np.round(intersection, 8)
            if ef_min + 0.001 < x < ef_max - 0.001:
                inner_cross_points.append([x - base_e, y])
            elif y > large_minus_number + 1:
                boundary_points.append([x - base_e, y])

        return CrossPoints(inner_cross_points, boundary_points)

    def transition_levels(
            self, base_e: float = 0.0) -> Dict[Tuple[int, int], float]:
        result = {}
        for (c1, e1), (c2, e2) in combinations(
                zip(self.charges, self.corrected_energies), 2):
            result[(c1, c2)] = - (e1 - e2) / (c1 - c2) - base_e
        return result

    def pinning_level(self,
                      base_e: float = 0.0
                      ) -> Tuple[Tuple[float, Optional[int]],
                                 Tuple[float, Optional[int]]]:
        """
        :param base_e: Reference to show the pinning level such as VBM.
        :return: ((Lower pinning, its charge), (Upper pinning, its charge))
        """
        lower_pinning, upper_pinning = float("-inf"), float("inf")
        lower_charge, upper_charge = None, None
        for charge, corr_energy in zip(self.charges, self.corrected_energies):
            if charge == 0:
                continue
            pinning = - corr_energy / charge
            if charge > 0 and pinning > lower_pinning:
                lower_pinning, lower_charge = pinning, charge
            elif pinning < upper_pinning:
                upper_pinning, upper_charge = pinning, charge
        return ((lower_pinning - base_e, lower_charge),
                (upper_pinning - base_e, upper_charge))

    def energy_at_ef(self, ef: float) -> Tuple[float, int]:
        """
        :return: (Lowest energy, its charge)
        """
        result_e, result_charge = float("inf"), None
        for charge, corr_energy in zip(self.charges, self.corrected_energies):
            energy = corr_energy + charge * ef
            if energy < result_e:
                result_e, result_charge = energy, charge
        return result_e, result_charge

    def __str__(self):
        lines = []
        for c, e, cor in zip(self.charges, self.energies, self.corrections):
            lines.append(f"{self.name:>10} {c:>4} {e:12.4f} {cor:12.4f}")
        return "\n".join(lines)


@dataclass
class CrossPoints:
    inner_cross_points: List[List[float]]
    boundary_points: List[List[float]]

    @property
    def all_sorted_points(self):
        return sorted(self.boundary_points + self.inner_cross_points,
                      key=lambda v: v[0])

    @property
    def t_all_sorted_points(self):
        return np.transpose(np.array(self.all_sorted_points)).tolist()

    @property
    def t_inner_cross_points(self):
        return np.transpose(np.array(self.inner_cross_points)).tolist()

    @property
    def t_boundary_points(self):
        return np.transpose(np.array(self.boundary_points)).tolist()

    @property
    def charges(self):
        result = []
        for i, j in zip(self.all_sorted_points[:-1], self.all_sorted_points[1:]):
            dx = j[0] - i[0]
            dy = j[1] - i[1]
            result.append(int(round(dy / dx)))
        return result

    @property
    def charge_list(self):
        charges = [None] + self.charges + [None]
        return list(zip(charges[:-1], charges[1:]))

    @property
    def annotated_charge_positions(self):
        result = {}
        for ((x1, y1), (x2, y2)), charge \
                in zip(zip(self.all_sorted_points[:-1],
                           self.all_sorted_points[1:]),
                       self.charges):
            result[charge] = [(x1 + x2) / 2, (y1 + y2) / 2]
        return result

    def __str__(self):
        lines = []
        for point in self.all_sorted_points:
            lines.append(f"{point[0]:12.4f} {point[1]:12.4f}")
        return "\n".join(lines)


def make_defect_energies(single_energies: List[SingleDefectEnergy]
                         ) -> List[DefectEnergy]:
    sorted_energies = sorted(single_energies, key=lambda x: x.name)
    result = []
    for _, grouped_energies in groupby(sorted_energies, lambda x: x.name):
        charges = []
        energies = []
        corrections = []
        for single_energy in grouped_energies:
            charges.append(single_energy.charge)
            energies.append(single_energy.energy)
            corrections.append(single_energy.correction)
        result.append(DefectEnergy(single_energy.name, charges, energies, corrections))
    return result


def make_energies(perfect: CalcResults,
                  defects: List[CalcResults],
                  defect_entries: List[DefectEntry],
                  corrections: List[Correction],
                  abs_chem_pot: Dict[Element, float],
                  allow_shallow: bool = True,
                  band_edges: List[BandEdgeStates] = None):

    single_energies = []
    for i, (d, e, c) in enumerate(zip(defects, defect_entries, corrections)):
        if allow_shallow is False:
            if band_edges and band_edges[i].is_shallow:
                continue
        single_energies.append(
            make_single_defect_energy(perfect, d, e, abs_chem_pot, c))
    return make_defect_energies(single_energies)


def make_single_defect_energy(perfect: CalcResults,
                              defect: CalcResults,
                              defect_entry: DefectEntry,
                              abs_chem_pot: Dict[Element, float],
                              correction: Optional[Correction] = NoCorrection()
                              ) -> SingleDefectEnergy:
    n_diffs = num_atom_differences(defect.structure, perfect.structure)
    energy = (defect.energy - perfect.energy
              + reservoir_energy(n_diffs, abs_chem_pot))
    return SingleDefectEnergy(defect_entry.name, defect_entry.charge, energy,
                              correction.correction_energy)


def reservoir_energy(diffs: Dict[Element, int],
                     abs_chem_pot: Dict[Element, float]) -> float:
    return sum([-diff * abs_chem_pot[elem] for elem, diff in diffs.items()])


def num_atom_differences(structure: IStructure,
                         ref_structure: IStructure) -> Dict[Element, int]:
    target_composition = structure.composition.as_dict()
    reference_composition = ref_structure.composition.as_dict()
    result = {}
    for k in set(target_composition.keys()) | set(reference_composition.keys()):
        n_atom_diff = int(target_composition[k] - reference_composition[k])
        if n_atom_diff:
            result[Element(k)] = n_atom_diff
    return result


def remove_digits(name):
    return ''.join([i for i in name if not i.isdigit()])


def only_digits(name):
    return ''.join([i for i in name if i.isdigit()])


elements = [str(e) for e in Element]


def defect_mpl_name(name):
    in_name, out_name = name.split("_")
    if in_name in elements:
        in_name = "{\\rm " + in_name + "}"
    elif in_name == "Va":
        in_name = "V"

    r_out_name = remove_digits(out_name)
    if r_out_name in elements:
        out_name = "{{\\rm " + r_out_name + "}" + only_digits(out_name) + "}"
    else:
        out_name = "{" + out_name + "}"

    return f"${in_name}_{out_name}$"


def defect_plotly_name(name):
    in_name, out_name = name.split("_")
    if in_name == "Va":
        in_name = "<i>V</i>"

    out_name = f"<sub>{out_name}</sub>"
    return f"{in_name}{out_name}"


def defect_plotly_full_name(fullname):
    in_name, out_name, charge = fullname.split("_")
    if in_name == "Va":
        in_name = "<i>V</i>"
    return f"{in_name}<sub>{out_name}</sub><sup>{charge}</sup>"


def defect_html_title_name(fullname):
    x = fullname.split("_")
    if len(x) == 2:
        in_name, out_name = x
    elif len(x) == 3:
        in_name, out_name, charge = x
    else:
        raise ValueError

    if in_name == "Va":
        in_name = html.I("V")
    else:
        in_name = html.Span(in_name)

    result = [in_name, html.Sub(out_name)]
    if len(x) == 3:
        result.append(html.Sup(charge))
    return result


def sanitize_defect_energies_for_plot(defect_energies: List[DefectEnergy],
                                      for_plotly: bool = False):
    result = []
    out_names = [e.name.split("_")[1] for e in defect_energies]

    for e in defect_energies:
        ee = deepcopy(e)
        in_name, out_name = e.name.split("_")
        r_out_name = remove_digits(out_name)
        out_name = r_out_name if f"{r_out_name}2" not in out_names else out_name
        if for_plotly:
            ee.name = defect_plotly_name("_".join([in_name, out_name]))
        else:
            ee.name = defect_mpl_name("_".join([in_name, out_name]))
        result.append(ee)

    return result


def slide_energy(defect_energies: List[DefectEnergy], base_level: float):
    result = []
    for e in defect_energies:
        ee = deepcopy(e)
        ee.energies = [e + base_level * c for e, c in zip(e.energies, e.charges)]
        result.append(ee)
    return result



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

from monty.serialization import loadfn
from pydefect.analyzer.band_edge_states import BandEdgeStates
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.dash_components.cpd_energy_dash import \
    CpdEnergyComponent, make_energies
from pydefect.analyzer.defect_energy import slide_energy
from pydefect.analyzer.defect_energy_plotter import DefectEnergyMplPlotter
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator
from pydefect.analyzer.defect_structure_info import make_defect_structure_info
from pydefect.analyzer.eigenvalue_plotter import EigenvalueMplPlotter
from pydefect.analyzer.make_band_edge_state import make_band_edge_state
from pydefect.chem_pot_diag.chem_pot_diag import ChemPotDiag, CpdPlotInfo, \
    CompositionEnergy, replace_comp_energy
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiagMpl2DMplPlotter, \
    ChemPotDiagMpl3DMplPlotter
from pydefect.chem_pot_diag.make_chem_pot_diag import make_chem_pot_diag_from_mp
from pydefect.cli.main_tools import sanitize_matrix
from pydefect.cli.vasp.make_band_edge_eigenvalues import \
    make_band_edge_eigenvalues
from pydefect.cli.vasp.make_calc_results import make_calc_results_from_vasp
from pydefect.cli.vasp.make_edge_characters import MakeEdgeCharacters
from pydefect.cli.vasp.make_efnv_correction import \
    make_efnv_correction
from pydefect.cli.vasp.make_gkfo_correction import make_gkfo_correction
from pydefect.cli.vasp.make_poscars_from_query import make_poscars_from_query
from pydefect.cli.vasp.make_unitcell import make_unitcell_from_vasp
from pydefect.corrections.site_potential_plotter import \
    SitePotentialMplPlotter
from pydefect.defaults import defaults
from pydefect.input_maker.add_interstitial import append_interstitial
from pydefect.input_maker.defect_entries_maker import DefectEntriesMaker
from pydefect.input_maker.defect_entry import DefectEntry
from pydefect.input_maker.defect_set import DefectSet
from pydefect.input_maker.defect_set_maker import DefectSetMaker
from pydefect.input_maker.supercell_info import SupercellInfo
from pydefect.input_maker.supercell_maker import SupercellMaker
from pydefect.util.mp_tools import MpQuery
from pymatgen.io.vasp import Vasprun, Outcar, Procar
from pymatgen.util.string import latexify
from vise.util.logger import get_logger

logger = get_logger(__name__)


def print_file(args):
    print(args.obj)


def make_unitcell(args):
    unitcell = make_unitcell_from_vasp(
        vasprun_band=args.vasprun_band,
        outcar_band=args.outcar_band,
        outcar_dielectric_clamped=args.outcar_dielectric_clamped,
        outcar_dielectric_ionic=args.outcar_dielectric_ionic)
    unitcell.to_json_file()


def make_competing_phase_dirs(args):
    query = MpQuery(element_list=args.elements, e_above_hull=args.e_above_hull)
    make_poscars_from_query(materials_query=query.materials, path=Path.cwd())


def make_chem_pot_diag(args) -> None:
    if args.elements:
        cpd = make_chem_pot_diag_from_mp(target=args.target,
                                         additional_elements=args.elements,
                                         vertex_elements=args.vertices,
                                         atom_energy_yaml=args.atom_energy_yaml)
    else:
        comp_es = []
        for d in args.dirs:
            vasprun = Vasprun(d / defaults.vasprun)
            composition = vasprun.final_structure.composition
            energy = float(vasprun.final_energy)  # type is FloatWithUnit
            comp_es.append(CompositionEnergy(composition, energy, "local"))
        if args.update:
            cpd = ChemPotDiag.from_yaml(args.yaml)
            replace_comp_energy(cpd, comp_es)
        else:
            cpd = ChemPotDiag(comp_es, args.target,
                              vertex_elements=args.vertices)
    cpd.to_yaml(args.yaml)


def plot_chem_pot_diag(args) -> None:
    cpd = ChemPotDiag.from_yaml(args.yaml)
    if cpd.dim == 1:
        logger.warning("Single element is not supported for the plot.")
        return

    print(cpd)
    if cpd.dim == 2:
        plotter = ChemPotDiagMpl2DMplPlotter(CpdPlotInfo(cpd))
    elif cpd.dim == 3:
        plotter = ChemPotDiagMpl3DMplPlotter(CpdPlotInfo(cpd))
    else:
        logger.info("Number of elements must be 2 or 3. "
                    f"Now, elements are {cpd.vertex_elements}.")
        return
    plt = plotter.draw_diagram()
    plt.savefig(fname="cpd.pdf")
    plt.show()


def make_supercell(args):
    if args.matrix:
        matrix = sanitize_matrix(args.matrix)
        maker = SupercellMaker(args.unitcell, matrix)
    else:
        kwargs = {}
        if args.min_num_atoms:
            kwargs["min_num_atoms"] = args.min_num_atoms
        if args.max_num_atoms:
            kwargs["max_num_atoms"] = args.max_num_atoms
        maker = SupercellMaker(args.unitcell, **kwargs)

    maker.supercell.structure.to(filename="SPOSCAR")
    maker.supercell_info.to_json_file()


def append_interstitial_to_supercell_info(args):
    supercell_info: SupercellInfo = append_interstitial(args.supercell_info,
                                                        args.base_structure,
                                                        args.frac_coords)
    supercell_info.to_json_file()


def pop_interstitial_from_supercell_info(args):
    supercell_info = args.supercell_info
    assert args.index > 0
    supercell_info.interstitials.pop(args.index - 1)
    supercell_info.to_json_file()


def make_defect_set(args):
    supercell_info = loadfn("supercell_info.json")
    if args.oxi_states:
        oxi_states = dict(zip(args.oxi_states[::2], args.oxi_states[1::2]))
    else:
        oxi_states = None
    maker = DefectSetMaker(supercell_info,
                           oxi_states,
                           args.dopants,
                           keywords=args.kwargs)
    maker.defect_set.to_yaml()


def make_defect_entries(args):
    supercell_info: SupercellInfo = loadfn("supercell_info.json")
    perfect = Path("perfect")

    try:
        perfect.mkdir()
        logger.info("Making perfect dir...")
        supercell_info.structure.to(filename=perfect / "POSCAR")
    except FileExistsError:
        logger.info(f"perfect dir exists, so skipped...")

    defect_set = DefectSet.from_yaml()
    maker = DefectEntriesMaker(supercell_info, defect_set)

    for defect_entry in maker.defect_entries:
        dir_path = Path(defect_entry.full_name)
        try:
            dir_path.mkdir()
            logger.info(f"Making {dir_path} dir...")
            defect_entry.perturbed_structure.to(filename=dir_path / "POSCAR")
            defect_entry.to_json_file(filename=dir_path / "defect_entry.json")
            defect_entry.to_prior_info(filename=dir_path / "prior_info.yaml")
        except FileExistsError:
            logger.info(f"{dir_path} dir exists, so skipped...")


def make_calc_results(args):
    for d in args.dirs:
        logger.info(f"Parsing data in {d} ...")
        calc_results = make_calc_results_from_vasp(
            vasprun=Vasprun(d / defaults.vasprun),
            outcar=Outcar(d / defaults.outcar))
        calc_results.to_json_file(filename=Path(d) / "calc_results.json")


def make_efnv_correction_from_vasp(args):
    for d in args.dirs:
        logger.info(f"Parsing data in {d} ...")
        defect_entry: DefectEntry = loadfn(d / "defect_entry.json")
        calc_results = loadfn(d / "calc_results.json")
        efnv = make_efnv_correction(defect_entry.charge,
                                    calc_results,
                                    args.perfect_calc_results,
                                    args.unitcell.dielectric_constant)
        efnv.to_json_file(d / "correction.json")

        title = defect_entry.full_name
        plotter = SitePotentialMplPlotter.from_efnv_corr(
            title=title, efnv_correction=efnv)
        plotter.construct_plot()
        plotter.plt.savefig(fname=d / "correction.pdf")
        plotter.plt.clf()


def make_gkfo_correction_from_vasp(args):
    gkfo = make_gkfo_correction(
        efnv_correction=args.initial_efnv_correction,
        additional_charge=args.charge_diff,
        final_calc_results=args.final_calc_results,
        initial_calc_results=args.initial_calc_results,
        diele_tensor=args.unitcell.dielectric_constant,
        ion_clamped_diele_tensor=args.unitcell.ele_dielectric_const)
    print(gkfo)
    gkfo.to_json_file("gkfo_correction.json")
    plotter = SitePotentialMplPlotter.from_gkfo_corr(
        title="GKFO correction", gkfo_correction=gkfo)
    plotter.construct_plot()
    plotter.plt.savefig(fname="gkfo_correction.pdf")
    plotter.plt.clf()


def make_defect_eigenvalues(args):
    supercell_vbm = args.perfect_calc_results.vbm
    supercell_cbm = args.perfect_calc_results.cbm
    for d in args.dirs:
        logger.info(f"Parsing data in {d} ...")
        try:
            defect_entry = loadfn(d / "defect_entry.json")
            title = defect_entry.name
        except FileNotFoundError:
            title = "No name"
        vasprun = Vasprun(d / defaults.vasprun)
        band_edge_eigenvalues = make_band_edge_eigenvalues(
            vasprun, supercell_vbm, supercell_cbm)
        band_edge_eigenvalues.to_json_file(d / "band_edge_eigenvalues.json")
        plotter = EigenvalueMplPlotter(
            title=title, band_edge_eigenvalues=band_edge_eigenvalues,
            supercell_vbm=supercell_vbm, supercell_cbm=supercell_cbm)
        plotter.construct_plot()
        plotter.plt.savefig(fname=d / "eigenvalues.pdf")
        plotter.plt.clf()


def make_edge_characters(args):
    for d in args.dirs:
        logger.info(f"Parsing data in {d} ...")
        vasprun = Vasprun(d / defaults.vasprun)
        procar = Procar(d / defaults.procar)
        outcar = Outcar(d / defaults.outcar)
        calc_results = loadfn(d / "calc_results.json")
        structure_comparator = DefectStructureComparator(
            calc_results.structure, args.perfect_calc_results.structure)
        edge_characters = MakeEdgeCharacters(
            procar, vasprun, outcar,
            structure_comparator.neighboring_atom_indices()).edge_characters
        edge_characters.to_json_file(d / "edge_characters.json")


def make_edge_states(args):
    for d in args.dirs:
        print(f"-- {d}")
        edge_states = []
        edge_characters = loadfn(d / "edge_characters.json")
        for spin, edge_character, ref in zip(["spin up  ", "spin down"],
                                             edge_characters,
                                             args.perfect_edge_characters):
            edge_state = make_band_edge_state(edge_character, ref)
            edge_states.append(edge_state)
            print(spin, edge_state)

        BandEdgeStates(edge_states).to_yaml(d / "band_edge_states.yaml")


def make_defect_formation_energy(args):
    formula = args.perfect_calc_results.structure.composition.reduced_formula
    chem_pot_diag = ChemPotDiag.from_yaml(args.cpd_yaml)
    pcr = args.perfect_calc_results

    defects, defect_entries, corrections, edge_states = [], [], [], []
    for d in args.dirs:
        if args.skip_shallow:
            edge_states = BandEdgeStates.from_yaml(d / "band_edge_states.yaml")
            if edge_states.is_shallow:
                continue
        defects.append(loadfn(d / "calc_results.json"))
        defect_entries.append(loadfn(d / "defect_entry.json"))
        corrections.append(loadfn(d / "correction.json"))

    if args.web_gui:
        from crystal_toolkit.settings import SETTINGS
        import dash_html_components as html
        from crystal_toolkit.helpers.layouts import Column
        import crystal_toolkit.components as ctc
        import dash

        edge_states = []
        for d in args.dirs:
            edge_states.append(BandEdgeStates.from_yaml(d / "band_edge_states.yaml"))

        app = dash.Dash(__name__,
                        suppress_callback_exceptions=True,
                        assets_folder=SETTINGS.ASSETS_PATH,
                        external_stylesheets=['https://codepen.io/chriddyp/pen/bWLwgP.css'])

        cpd_plot_info = CpdPlotInfo(chem_pot_diag)
        cpd_e_component = CpdEnergyComponent(cpd_plot_info,
                                             pcr,
                                             defects,
                                             defect_entries,
                                             corrections,
                                             args.unitcell.vbm,
                                             args.unitcell.cbm,
                                             edge_states)
        my_layout = html.Div([Column(cpd_e_component.layout)])
        ctc.register_crystal_toolkit(app=app, layout=my_layout, cache=None)
        app.run_server(port=args.port)
        return

    abs_chem_pot = chem_pot_diag.abs_chem_pot_dict(args.label)
    title = " ".join([latexify(formula), "point", args.label])
    defect_energies = make_energies(pcr, defects, defect_entries, corrections, abs_chem_pot)

    if args.print:
        if args.supercell_edge:
            vbm, cbm = pcr.vbm, pcr.cbm
        else:
            vbm, cbm = args.unitcell.vbm, args.unitcell.cbm

        defect_energies = slide_energy(defect_energies, vbm)
        print("         charge          E_f   correction    ")
        for e in defect_energies:
            print(e)
            print("")

        print("-- cross points -- ")
        for e in defect_energies:
            print(e.name)
            print(e.cross_points(ef_min=0, ef_max=cbm - vbm))
            print("")
        return

    plotter = DefectEnergyMplPlotter(title=title,
                                     defect_energies=defect_energies,
                                     vbm=args.unitcell.vbm,
                                     cbm=args.unitcell.cbm,
                                     supercell_vbm=pcr.vbm,
                                     supercell_cbm=pcr.cbm,
                                     y_range=args.y_range,
                                     supercell_edge=args.supercell_edge,
                                     label_line=args.label_line,
                                     add_charges=args.add_charges)

    plotter.construct_plot()
    plotter.plt.savefig(f"energy_{args.label}.pdf")


def calc_defect_structure_info(args):
    supercell_info: SupercellInfo = args.supercell_info
    for d in args.dirs:
        logger.info(f"Parsing data in {d} ...")
        calc_results: CalcResults = loadfn(d / "calc_results.json")
        defect_entry: DefectEntry = loadfn(d / "defect_entry.json")
        defect_str_info = make_defect_structure_info(
            supercell_info.structure,
            defect_entry.structure,
            calc_results.structure,
            dist_tol=args.dist_tolerance,
            symprec=args.symprec,
            init_site_sym=defect_entry.site_symmetry,
            final_site_sym=calc_results.site_symmetry)
        defect_str_info.to_json_file(str(d / "defect_structure_info.json"))



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pydefect.corrections.abstract_correction import Correction


class NoCorrection(Correction):
    @property
    def correction_energy(self):
        return 0.0


class ManualCorrection(Correction):
    def __init__(self, correction_energy: float):
        self._correction_energy = correction_energy

    @property
    def correction_energy(self):
        return self._correction_energy



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import fire
from pydefect.input_maker.defect_entry import make_defect_entry
from pymatgen.core import IStructure
from vise.util.logger import get_logger

logger = get_logger(__name__)


def make_defect_entry_from_poscars(original_name, perf_poscar, defect_poscar):
    names = original_name.split("_")
    try:
        name = "_".join(names[:-1])
        charge = int(names[-1])
    except ValueError:
        logger.warning(f"Name {original_name} is invalid.")
        raise

    perfect = IStructure.from_file(perf_poscar)
    defect = IStructure.from_file(defect_poscar)

    defect_entry = make_defect_entry(name=name,
                                     charge=charge,
                                     perfect_structure=perfect,
                                     defect_structure=defect)
    defect_entry.to_json_file()


if __name__ == '__main__':
    fire.Fire(make_defect_entry_from_poscars)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest

from pydefect.analyzer.unitcell import Unitcell
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip


@pytest.fixture
def unitcell():
    ele = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    ion = [[10, 20, 30], [40, 50, 60], [70, 80, 90]]
    return Unitcell(vbm=0.1, cbm=5.1,
                    ele_dielectric_const=ele, ion_dielectric_const=ion)


def test_unitcell(unitcell):
    expected = np.array([[11, 22, 33], [44, 55, 66], [77, 88, 99]])
    np.testing.assert_array_almost_equal(unitcell.dielectric_constant, expected)


def test_msonable(unitcell):
    assert_msonable(unitcell)


def test_json_round_trip(unitcell, tmpdir):
    assert_json_roundtrip(unitcell, tmpdir)



# -*- coding: utf-8 -*-

from typing import List

from pymatgen.ext.matproj import MPRester
from pymatgen.core import Element

from pydefect.defaults import defaults

elements = [e.name for e in Element]


class MpQuery:
    def __init__(self,
                 element_list: List[str],
                 e_above_hull: float = defaults.e_above_hull,
                 properties: List[str] = None,
                 ):

        self.element_list = element_list
        self.properties = (properties or
                           ["task_id", "full_formula", "final_energy",
                            "structure", "spacegroup", "band_gap",
                            "total_magnetization", "magnetic_type"])
        self.e_above_hull = e_above_hull

        excluded = list(set(elements) - set(element_list))
        # API key is parsed via .pmgrc.yaml

        with MPRester() as m:
            # Due to mp_decode=True by default, class objects are restored.
            self.materials = \
                m.query(criteria={"elements": {"$in": element_list,
                                               "$nin": excluded},
                                  "e_above_hull": {"$lte": e_above_hull}},
                        properties=self.properties)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

from typing import List

import dash_table
from crystal_toolkit.core.mpcomponent import MPComponent
from crystal_toolkit.helpers.layouts import Column, dcc, html, Columns
from dash.dependencies import Input, Output
from pydefect.analyzer.band_edge_states import BandEdgeStates
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_energy import make_energies
from pydefect.analyzer.defect_energy_plotter import DefectEnergyPlotlyPlotter
from pydefect.chem_pot_diag.chem_pot_diag import CpdPlotInfo
from pydefect.chem_pot_diag.cpd_plotter import ChemPotDiagPlotly2DMplPlotter, \
    ChemPotDiagPlotly3DMplPlotter
from pydefect.corrections.abstract_correction import Correction
from pydefect.input_maker.defect_entry import DefectEntry


class CpdEnergyComponent(MPComponent):
    def __init__(self,
                 cpd_plot_info: CpdPlotInfo,
                 perfect: CalcResults,
                 defect_calc_results: List[CalcResults],
                 defect_entries: List[DefectEntry],
                 corrections: List[Correction],
                 unitcell_vbm: float,
                 unitcell_cbm: float,
                 band_edge_states: List[BandEdgeStates] = None,
                 show_competing_phases = False,
                 *args, **kwargs):
        self.cpd_plot_info = cpd_plot_info
        self.perfect = perfect
        self.defect_calc_results = defect_calc_results
        self.defect_entries = defect_entries
        self.corrections = corrections
        self.unitcell_vbm = unitcell_vbm
        self.unitcell_cbm = unitcell_cbm
        self.band_edge_states = band_edge_states
        self.dim = cpd_plot_info.dim
        super().__init__(*args, **kwargs)

        df = self.cpd_plot_info.cpd.target_vertex_list_dataframe.copy()
        df.reset_index(inplace=True)

        def rename(x):
            if "mu" in x:
                return x.split("_")[1]
            return x

        if show_competing_phases:
            columns = [{"name": rename(i), "id": i} for i in df.columns]
        else:
            columns = [{"name": rename(i), "id": i}
                       for i in df.columns if "competing phase" not in i]

        columns[0]["name"] = "label"

        self.vertex_list = dash_table.DataTable(
            id='table',
            columns=columns,
            data=df.to_dict('records'),
            style_cell={'fontSize': 24,
                        'font-family': 'sans-serif',
                        "width": 700})

    def energy_fig(self, abs_chem_pot, label, allow_shallow):
        defect_energies = make_energies(self.perfect,
                                        self.defect_calc_results,
                                        self.defect_entries,
                                        self.corrections,
                                        abs_chem_pot,
                                        allow_shallow,
                                        self.band_edge_states)
        plotter = DefectEnergyPlotlyPlotter(title=f"Label {label[0]}",
                                            defect_energies=defect_energies,
                                            vbm=self.unitcell_vbm,
                                            cbm=self.unitcell_cbm,
                                            supercell_vbm=self.perfect.vbm,
                                            supercell_cbm=self.perfect.cbm,
                                            supercell_edge=False)
        return plotter.create_figure()

    @property
    def _sub_layouts(self):

        allow_shallow = self.get_choice_input(
            kwarg_label="allow_shallow",
            state={"label": "True", "value": "True"},
            label="Allow shallow states",
            # Only string values are allowed.
            options=[{"label": "True", "value": "True"},
                     {"label": "False", "value": "False"}])

        if self.dim in [2, 3]:
            if self.cpd_plot_info.dim == 2:
                fig = ChemPotDiagPlotly2DMplPlotter(self.cpd_plot_info).figure
            else:
                fig = ChemPotDiagPlotly3DMplPlotter(self.cpd_plot_info).figure
            cpd = dcc.Graph(id="cpd", figure=fig)
            energy = dcc.Graph(id=self.id('energy'))
            return {"cpd": cpd, "energy": energy, "allow_shallow": allow_shallow}
        else:
            d = [{"label": v, "value": v}
                 for v in self.cpd_plot_info.cpd.target_vertices.keys()]
            cpd_label = self.get_choice_input(
                kwarg_label="label",
                state={"label": "A", "value": "A"},
                label="Equilibrium label",
                options=d)
            energy = dcc.Graph(id=self.id('energy'))
            return {"cpd_label": cpd_label, "energy": energy, "allow_shallow": allow_shallow}

    @property
    def layout(self):
        if self.dim in [2, 3]:
            return Columns([Column([html.Div(self._sub_layouts["energy"]),
                                    self._sub_layouts["allow_shallow"]]),
                            Column([self._sub_layouts["cpd"],
                                    self.vertex_list])
                            ])

        else:
            return Columns([Column([self.vertex_list,
                                    self._sub_layouts["allow_shallow"],
                                    self._sub_layouts["cpd_label"]]),
                            Column(html.Div(self._sub_layouts["energy"]))])

    def generate_callbacks(self, app, cache):
        if self.dim in [2, 3]:
            @app.callback(
                Output(self.id('energy'), 'figure'),
                [Input('cpd', 'clickData'),
                 Input(self.get_kwarg_id('allow_shallow'), 'value')])
            def display_cpd(click_data, allow_shallow):
                allow_shallow = True if allow_shallow[0] == "True" else False
                label = click_data["points"][0]["text"] if click_data else "A"
                abs_chem_pot = self.cpd_plot_info.cpd.abs_chem_pot_dict(label)
                return self.energy_fig(abs_chem_pot, label, allow_shallow)
        else:
            @app.callback(
                Output(self.id('energy'), 'figure'),
                [Input(self.get_kwarg_id("label"), "value"),
                 Input(self.get_kwarg_id('allow_shallow'), 'value')])
            def display_cpd(label, allow_shallow):
                allow_shallow = True if allow_shallow[0] == "True" else False
                label = label[0] if label else "A"
                abs_chem_pot = self.cpd_plot_info.cpd.abs_chem_pot_dict(label)
                return self.energy_fig(abs_chem_pot, label, allow_shallow)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from collections import defaultdict
from itertools import zip_longest

from pydefect.analyzer.band_edge_states import EdgeState, EdgeCharacter
from pydefect.defaults import defaults


def make_band_edge_state(target: EdgeCharacter, ref: EdgeCharacter) -> EdgeState:
    is_hob_near_cbm = (target.hob_bottom_e
                       > ref.cbm - defaults.similar_energy_criterion)
    is_lub_near_vbm = (target.lub_top_e
                       < ref.vbm + defaults.similar_energy_criterion)

    if target.vbm is not None:
        similar_vbm_e = (abs(target.vbm - ref.vbm)
                         < defaults.similar_energy_criterion)
    else:
        similar_vbm_e = False

    if target.cbm is not None:
        similar_cbm_e = (abs(target.cbm - ref.cbm)
                         < defaults.similar_energy_criterion)
    else:
        similar_cbm_e = False

    similar_vbm_o = are_orbitals_similar(target.vbm_orbitals, ref.vbm_orbitals)
    similar_cbm_o = are_orbitals_similar(target.cbm_orbitals, ref.cbm_orbitals)

    if similar_vbm_e and similar_cbm_e and similar_vbm_o and similar_cbm_o:
        return EdgeState.no_in_gap
    elif is_hob_near_cbm:
        return EdgeState.donor_phs
    elif is_lub_near_vbm:
        return EdgeState.acceptor_phs
    else:
        return EdgeState.in_gap_state


def are_orbitals_similar(orbital_1: dict, orbital_2: dict) -> float:
    element_set = set(list(orbital_1.keys()) + list(orbital_2.keys()))
    orb_1, orb_2 = defaultdict(list, orbital_1), defaultdict(list, orbital_2)

    difference = 0
    for e in element_set:
        difference += sum([abs(i - j) for i, j
                           in zip_longest(orb_1[e], orb_2[e], fillvalue=0)])

    return difference < defaults.similar_orb_criterion



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest
from pymatgen.core import Structure, Lattice
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Vasprun, Outcar, Procar

from pydefect.cli.vasp.make_edge_characters import MakeEdgeCharacters, \
    calc_participation_ratio, calc_orbital_character


@pytest.fixture
def make_edge_characters(mocker):
    mock_procar = mocker.Mock(spec=Procar, autospec=True)
    mock_vasprun = mocker.Mock(spec=Vasprun, autospec=True)
    mock_outcar = mocker.Mock(spec=Outcar, autospec=True)
#    mock_ratio = mocker.patch("pydefect.cli.vasp.make_edge_characters.calc_participation_ratio")
    mock_procar.data = {Spin.up: np.array(
        # 1st atom
        # first k-index
        [[[[0.0] * 15 + [0.1]] * 3,   # 1st band
          [[0.1] * 16, [0.2] * 16, [1.0] * 16],
          [[0.0] * 15 + [0.1]] * 3],  # 2nd band
         # 2nd k-index      2nd k-index
         [[[0.0] * 15 + [0.1]] * 3,
          [[0.0] * 15 + [0.1]] * 3,
          [[0.0] * 15 + [0.1]] * 3]]),
    Spin.down: np.array(
        # 1st atom
        # first k-index
        [[[[0.0] * 15 + [0.1]] * 3,   # 1st band
          [[0.1] * 16, [0.2] * 16, [1.0] * 16],
          [[0.0] * 15 + [0.1]] * 3],  # 2nd band
         # 2nd k-index      2nd k-index
         [[[0.0] * 15 + [0.1]] * 3,
          [[0.0] * 15 + [0.1]] * 3,
          [[0.0] * 15 + [0.1]] * 3]])}
    mock_vasprun.final_structure = Structure(Lattice.cubic(1), species=["H", "H", "He"], coords=[[0] * 3] * 3)
    mock_vasprun.eigenvalues = {Spin.up: np.array([[[-3.01, 1.],
                                                    [-2.90, 1.],  # vbm
                                                    [ 8.01, 0.]],
                                                   [[-3.01, 1.],
                                                    [-2.99, 1.],
                                                    [ 7.90, 0.]]]), # cbm
                    # shift one band for testing magnetization
                                Spin.down: np.array([[[-2.99, 1.],
                                                      [ 7.99, 0.],
                                                      [10.00, 0.]],
                                                     [[-3.01, 1.],
                                                      [ 7.99, 0.],
                                                      [10.00, 0.]]])}
    mock_outcar.nelect = 3.0
    mock_outcar.total_mag = 1

    return MakeEdgeCharacters(mock_procar, mock_vasprun, mock_outcar, [0, 1])


def test_edge_characters(make_edge_characters):
    edge_characters = make_edge_characters.edge_characters
    assert edge_characters[0].vbm == -2.90
    assert edge_characters[0].hob_bottom_e == -2.99
    assert edge_characters[0].cbm == 7.90
    assert edge_characters[0].lub_top_e == 8.01

    assert edge_characters[1].vbm == -2.99
    assert edge_characters[1].hob_bottom_e == -3.01
    assert edge_characters[1].cbm == 7.99
    assert edge_characters[1].lub_top_e == 7.99


def test_calc_participation_ratio():
    # only a single atom
    actual = calc_participation_ratio(orbitals={Spin.up: np.array(
        # first k-index
        [[[[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]],   # 1st band
          [[0.2, 0.2, 0.2], [0.2, 0.2, 0.2], [0.3, 0.3, 0.3]]],  # 2nd band
         # 2nd k-index
         [[[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]],
          [[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]]]])},
        spin=Spin.up, kpt_index=0, band_index=1, atom_indices=[0, 1])
    np.testing.assert_almost_equal(actual, 1.2 / (1.2 + 0.9))


def test_calc_orbital_character_azimuthal_false():
    spd_orbitals = {Spin.up: np.array(
        # 1st atom
        # first k-index
        [[[[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]],   # 1st band
          [[0.1, 0.1, 0.1], [0.2, 0.3, 0.4], [1.0, 1.1, 1.2]]],  # 2nd band
         # 2nd k-index      2nd k-index
         [[[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]],
          [[0.0, 0.0, 0.1], [0.0, 0.0, 0.1], [0.0, 0.0, 0.1]]]])}

    actual = calc_orbital_character(
        orbitals=spd_orbitals,
        structure=Structure(Lattice.cubic(1),
                            species=["H", "H", "He"], coords=[[0] * 3] * 3),
        spin=Spin.up, kpt_index=0, band_index=1)

    assert actual == {"H": [0.3, 0.4, 0.5], "He": [1.0, 1.1, 1.2]}


def test_calc_orbital_character_azimuthal_true():
    azimuthal_orbitals = {Spin.up: np.array(
        # 1st atom
        # first k-index
        [[[[0.0] * 15 + [0.1]] * 3,   # 1st band
          [[0.1] * 16, [0.2] * 16, [1.0] * 16]],  # 2nd band
         # 2nd k-index      2nd k-index
         [[[0.0] * 15 + [0.1]] * 3,
          [[0.0] * 15 + [0.1]] * 3]])}

    actual = calc_orbital_character(
        orbitals=azimuthal_orbitals,
        structure=Structure(Lattice.cubic(1),
                            species=["H", "H", "He"], coords=[[0] * 3] * 3),
        spin=Spin.up, kpt_index=0, band_index=1)

    assert actual == {"H": [0.3, 0.9, 1.5, 2.1], "He": [1.0, 3.0, 5.0, 7.0]}


"""
TODO
-

DONE
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from argparse import Namespace
from pathlib import Path

import numpy as np
import pytest
from monty.serialization import loadfn
from pydefect.analyzer.band_edge_states import EdgeCharacters
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.unitcell import Unitcell
from pydefect.cli.vasp.main_function import make_supercell, make_defect_set, \
    make_defect_entries, make_unitcell, make_competing_phase_dirs, \
    make_chem_pot_diag, make_calc_results, print_file, \
    make_efnv_correction_from_vasp, make_defect_formation_energy, \
    make_defect_eigenvalues, make_edge_characters, \
    append_interstitial_to_supercell_info, pop_interstitial_from_supercell_info, \
    plot_chem_pot_diag, make_gkfo_correction_from_vasp, calc_defect_structure_info
from pydefect.corrections.efnv_correction import \
    ExtendedFnvCorrection
from pydefect.defaults import defaults
from pydefect.input_maker.defect import SimpleDefect
from pydefect.input_maker.defect_entry import DefectEntry
from pydefect.input_maker.defect_set import DefectSet
from pydefect.input_maker.supercell_info import SupercellInfo
from pymatgen.core import IStructure, Composition, Structure, Lattice, Element
from pymatgen.io.vasp import Vasprun, Outcar


def test_print():
    args = Namespace(obj="a")
    print_file(args)


def test_make_unitcell(mocker):
    vasprun_band_mock = mocker.Mock(spec=Vasprun, autospec=True)
    outcar_band_mock = mocker.Mock(spec=Outcar, autospec=True)
    outcar_dielectric_mock = mocker.Mock(spec=Outcar, autospec=True)
    args = Namespace(vasprun_band=vasprun_band_mock,
                     outcar_band=outcar_band_mock,
                     outcar_dielectric_clamped=outcar_dielectric_mock,
                     outcar_dielectric_ionic=outcar_dielectric_mock)

    mock = mocker.patch("pydefect.cli.vasp.main_function.make_unitcell_from_vasp")
    mock.return_value = Unitcell(vbm=1.0,
                                 cbm=2.0,
                                 ele_dielectric_const=np.eye(3),
                                 ion_dielectric_const=np.eye(3))
    make_unitcell(args)
    mock.assert_called_once_with(vasprun_band=vasprun_band_mock,
                                 outcar_band=outcar_band_mock,
                                 outcar_dielectric_clamped=outcar_dielectric_mock,
                                 outcar_dielectric_ionic=outcar_dielectric_mock)


def test_make_competing_phase_dirs(mocker):
    args = Namespace(elements=["Mg", "O"],
                     e_above_hull=0.1)
    mock = mocker.patch("pydefect.cli.vasp.main_function.MpQuery")
    mock_make = mocker.patch("pydefect.cli.vasp.main_function.make_poscars_from_query")
    make_competing_phase_dirs(args)
    mock.assert_called_once_with(element_list=args.elements,
                                 e_above_hull=args.e_above_hull)
    mock_make.assert_called_once_with(materials_query=mock.return_value.materials, path=Path.cwd())


def test_make_chem_pot_diag(mocker, tmpdir):
    def side_effect(key):
        mock_vasprun = mocker.Mock()
        if key == Path("Mg") / defaults.vasprun:
            mock_vasprun.final_structure.composition = Composition("Mg2")
            mock_vasprun.final_energy = -10
        elif key == Path("O") / defaults.vasprun:
            mock_vasprun.final_structure.composition = Composition("O2")
            mock_vasprun.final_energy = -20
        elif key == Path("MgO") / defaults.vasprun:
            mock_vasprun.final_structure.composition = Composition("MgO")
            mock_vasprun.final_energy = -30
        elif key == Path("Al") / defaults.vasprun:
            mock_vasprun.final_structure.composition = Composition("Al")
            mock_vasprun.final_energy = 0
        elif key == Path("MgAl2O4") / defaults.vasprun:
            mock_vasprun.final_structure.composition = Composition("MgAl2O4")
            mock_vasprun.final_energy = -70
        else:
            raise ValueError
        return mock_vasprun

    tmpdir.chdir()
    mock = mocker.patch("pydefect.cli.vasp.main_function.Vasprun", side_effect=side_effect)
    args_1 = Namespace(elements=None, functional=None, yaml="cpd.yaml", update=False,
                       dirs=[Path("Mg"), Path("MgO"), Path("O"), Path("Al"), Path("MgAl2O4")],
                       target=Composition("MgO"),
                       vertices=None)
    make_chem_pot_diag(args_1)

    args = Namespace(yaml="cpd.yaml")
    plot_chem_pot_diag(args)

    args_2 = Namespace(elements=None, functional=None, yaml="cpd.yaml", update=False,
                       dirs=[Path("Mg"), Path("MgO"), Path("O"), Path("Al"), Path("MgAl2O4")],
                       target=Composition("MgO"),
                       vertices=["Mg", "Al", "O"])
    make_chem_pot_diag(args_2)

    args = Namespace(yaml="cpd.yaml")
    plot_chem_pot_diag(args)  # show Al phases

    args_3 = Namespace(elements=None, functional=None, yaml="cpd.yaml", update=False,
                       dirs=[Path("Mg"), Path("MgO"), Path("O"), Path("Al"), Path("MgAl2O4")],
                       target=Composition("MgAl2O4"), vertices=None)
    make_chem_pot_diag(args_3)

    args = Namespace(yaml="cpd.yaml")
    plot_chem_pot_diag(args)



def test_make_supercell_from_matrix(simple_cubic, simple_cubic_2x1x1, tmpdir):
    matrix = [2, 1, 1]
    args = Namespace(unitcell=simple_cubic, matrix=matrix, min_num_atoms=None, max_num_atoms=None)

    tmpdir.chdir()
    make_supercell(args)
    info = loadfn("supercell_info.json")
    assert IStructure.from_file("SPOSCAR") == simple_cubic_2x1x1
    assert info.structure == simple_cubic_2x1x1
    assert info.transformation_matrix == [[2, 0, 0], [0, 1, 0], [0, 0, 1]]


def test_make_recommended_supercell(simple_cubic, simple_cubic_2x2x2, tmpdir):
    args = Namespace(unitcell=simple_cubic, matrix=None, min_num_atoms=8, max_num_atoms=8)

    tmpdir.chdir()
    make_supercell(args)
    info = loadfn("supercell_info.json")
    assert IStructure.from_file("SPOSCAR") == simple_cubic_2x2x2
    assert info.structure == simple_cubic_2x2x2
    assert info.transformation_matrix == [[2, 0, 0], [0, 2, 0], [0, 0, 2]]


def test_add_interstitials(mocker):
    mock_1 = mocker.Mock()
    mock_2 = mocker.Mock()
    mock_3 = mocker.Mock()
    args = Namespace(supercell_info=mock_1, base_structure=mock_2, frac_coords=mock_3)
    mock = mocker.patch("pydefect.cli.vasp.main_function.append_interstitial")

    append_interstitial_to_supercell_info(args)
    mock.assert_called_once_with(mock_1, mock_2, mock_3)
    mock.return_value.to_json_file.assert_called_once_with()


def test_pop_interstitials(mocker):
    mock_si = mocker.MagicMock()
    args = Namespace(supercell_info=mock_si, index=1000)

    pop_interstitial_from_supercell_info(args)
    mock_si.interstitials.pop.assert_called_once_with(999)
    mock_si.to_json_file.assert_called_once_with()


@pytest.mark.parametrize("oxi_states,he_vacancy_charge",
                         ([None, [0]], [["He", 1, "Li", 1], [-1, 0, 1]]))
def test_make_defect_set(oxi_states, he_vacancy_charge, tmpdir, supercell_info):
    tmpdir.chdir()
    supercell_info.to_json_file()
    args = Namespace(oxi_states=oxi_states, dopants=["Li"], kwargs=["Li_H1", "Va_He1", "Va_H1_-1"])
    make_defect_set(args)

    simple_defects = {SimpleDefect(None, "He1", he_vacancy_charge),
                      SimpleDefect(None, "H1", [-1]),
                      SimpleDefect("Li", "H1", [0])}

    DefectSet(defects=simple_defects).to_yaml("expected.yaml")
    assert Path("defect_in.yaml").read_text() == Path("expected.yaml").read_text()


def test_make_defect_entries(tmpdir, supercell_info):
    tmpdir.chdir()
    supercell_info.to_json_file()
    defect_set = DefectSet({SimpleDefect(None, "He1", [-1, 0])})
    defect_set.to_yaml()
    args = Namespace()
    make_defect_entries(args)
    names = {str(name) for name in Path(".").glob("*")}
    assert names == {'Va_He1_-1', 'defect_in.yaml', 'perfect', 'Va_He1_0', 'supercell_info.json'}

    perfect_structure = Structure.from_file(Path("perfect") / "POSCAR")
    assert perfect_structure == supercell_info.structure

    file_names = {str(file_name.name) for file_name in Path("Va_He1_-1").glob("*")}
    assert file_names == {"POSCAR", "defect_entry.json", "prior_info.yaml"}

    expected = """charge: -1
"""
    assert Path("Va_He1_-1/prior_info.yaml").read_text() == expected


def test_make_calc_results(tmpdir, mocker):
    tmpdir.chdir()
    mock = mocker.patch("pydefect.cli.vasp.main_function.make_calc_results_from_vasp")
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_function.Vasprun")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main_function.Outcar")
    mock_calc_results = mocker.Mock(spec=CalcResults)
    mock.return_value = mock_calc_results
    args = Namespace(dirs=[Path("a")])
    make_calc_results(args)

    mock_vasprun.assert_called_with(Path("a") / defaults.vasprun)
    mock_outcar.assert_called_with(Path("a") / defaults.outcar)
    mock.assert_called_with(vasprun=mock_vasprun.return_value, outcar=mock_outcar.return_value)
    mock_calc_results.to_json_file.assert_called_with(filename=Path("a") / "calc_results.json")


def test_make_efnv_correction_from_vasp(tmpdir, mocker):
    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    def side_effect(key):
        if str(key) == "Va_O1_2/defect_entry.json":
            mock_defect_entry.full_name = "Va_O1_2"
            mock_defect_entry.charge = 2
            return mock_defect_entry
        elif str(key) == "Va_O1_2/calc_results.json":
            return mock_calc_results
        else:
            raise ValueError

    mock_perfect_calc_results = mocker.Mock(spec=CalcResults)
    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_function.loadfn", side_effect=side_effect)
    mock_unitcell = mocker.Mock(spec=Unitcell)
    mock_make_efnv = mocker.patch("pydefect.cli.vasp.main_function.make_efnv_correction")
    mock_efnv = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_make_efnv.return_value = mock_efnv

    mock_pot_plotter = mocker.patch("pydefect.cli.vasp.main_function.SitePotentialMplPlotter")

    args = Namespace(dirs=[Path("Va_O1_2")],
                     perfect_calc_results=mock_perfect_calc_results,
                     unitcell=mock_unitcell)

    make_efnv_correction_from_vasp(args)
    mock_loadfn.assert_any_call(Path("Va_O1_2") / "defect_entry.json")
    mock_loadfn.assert_any_call(Path("Va_O1_2") / "calc_results.json")

    mock_make_efnv.assert_called_with(mock_defect_entry.charge, mock_calc_results, mock_perfect_calc_results, mock_unitcell.dielectric_constant)
    mock_efnv.to_json_file.assert_called_with(Path("Va_O1_2") / "correction.json")

    mock_pot_plotter.from_efnv_corr.assert_called_with(title="Va_O1_2", efnv_correction=mock_efnv)
    mock_pot_plotter.from_efnv_corr.return_value.construct_plot.assert_called_once_with()
    mock_pot_plotter.from_efnv_corr.return_value.plt.savefig.assert_called_once_with(fname=Path("Va_O1_2") / "correction.pdf")


def test_make_gkfo_correction_from_vasp(tmpdir, mocker):
    mock_i_correction = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)
    mock_i_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_f_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_unitcell = mocker.Mock()

    mock_pot_plotter = mocker.patch("pydefect.cli.vasp.main_function.SitePotentialMplPlotter")
    mock_make_gkfo = mocker.patch("pydefect.cli.vasp.main_function.make_gkfo_correction")

    args = Namespace(
        initial_efnv_correction=mock_i_correction,
        initial_calc_results=mock_i_calc_results,
        final_calc_results=mock_f_calc_results,
        charge_diff=1,
        unitcell=mock_unitcell)

    make_gkfo_correction_from_vasp(args)
    mock_make_gkfo.assert_called_with(
        efnv_correction=mock_i_correction,
        additional_charge=1,
        final_calc_results=mock_f_calc_results,
        initial_calc_results=mock_i_calc_results,
        diele_tensor=mock_unitcell.dielectric_constant,
        ion_clamped_diele_tensor=mock_unitcell.ele_dielectric_const)


def test_make_defect_eigenvalues(mocker):
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_function.Vasprun")

    mock_make_eigvals = mocker.patch(
        "pydefect.cli.vasp.main_function.make_band_edge_eigenvalues")

    mock_perfect_calc_results = mocker.Mock(spec=CalcResults)
    mock_perfect_calc_results.vbm = 10
    mock_perfect_calc_results.cbm = 20

    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)

    mock_eigval_plotter = mocker.patch(
        "pydefect.cli.vasp.main_function.EigenvalueMplPlotter")

    def side_effect(key):
        if str(key) == "Va_O1_2/defect_entry.json":
            mock_defect_entry.name = "Va_O1"
            mock_defect_entry.charge = 2
            return mock_defect_entry
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_function.loadfn",
                               side_effect=side_effect)

    args = Namespace(dirs=[Path("Va_O1_2")],
                     perfect_calc_results=mock_perfect_calc_results)
    make_defect_eigenvalues(args)

    mock_vasprun.assert_called_with(Path("Va_O1_2") / defaults.vasprun)
    mock_make_eigvals.assert_called_with(mock_vasprun.return_value, 10, 20)
    mock_make_eigvals.return_value.to_json_file.assert_called_with(
        Path("Va_O1_2") / "band_edge_eigenvalues.json")
    mock_loadfn.assert_any_call(Path("Va_O1_2") / "defect_entry.json")
    mock_eigval_plotter.assert_called_with(
        title="Va_O1",
        band_edge_eigenvalues=mock_make_eigvals.return_value,
        supercell_vbm=10,
        supercell_cbm=20)


def test_make_edge_characters(mocker):
    mock_vasprun = mocker.patch("pydefect.cli.vasp.main_function.Vasprun")
    mock_procar = mocker.patch("pydefect.cli.vasp.main_function.Procar")
    mock_outcar = mocker.patch("pydefect.cli.vasp.main_function.Outcar")

    mock_perfect_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_perfect_calc_results.structure = mocker.Mock(spec=Structure)
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)

    mock_analyzer = mocker.patch(
        "pydefect.cli.vasp.main_function.DefectStructureComparator")
    mock_characters = mocker.patch(
        "pydefect.cli.vasp.main_function.MakeEdgeCharacters")

    def side_effect(key):
        if str(key) == "Va_O1_2/calc_results.json":
            mock_calc_results.structure = mocker.Mock(spec=Structure, autospec=True)
            return mock_calc_results
        else:
            raise ValueError
    mocker.patch("pydefect.cli.vasp.main_function.loadfn", side_effect=side_effect)

    args = Namespace(dirs=[Path("Va_O1_2")],
                     perfect_calc_results=mock_perfect_calc_results)
    make_edge_characters(args)

    mock_vasprun.assert_called_with(Path("Va_O1_2") / defaults.vasprun)
    mock_procar.assert_called_with(Path("Va_O1_2") / defaults.procar)
    mock_outcar.assert_called_with(Path("Va_O1_2") / defaults.outcar)
    mock_analyzer.assert_called_with(mock_calc_results.structure,
                                     mock_perfect_calc_results.structure)
    mock_characters.assert_called_with(mock_procar.return_value,
                                       mock_vasprun.return_value,
                                       mock_outcar.return_value,
                                       mock_analyzer.return_value.neighboring_atom_indices.return_value)


def test_make_edge_state(mocker):
    mock_perf_edge_chars = mocker.Mock(spec=EdgeCharacters, autospec=True)
    args = Namespace(dirs=[Path("Va_O1_2")],
                     perfect_edge_characters=mock_perf_edge_chars)

    mock_edge_chars = mocker.Mock(spec=EdgeCharacters, autospec=True)
    def side_effect(key):
        if str(key) == "Va_O1_2/edge_characters.json":
            return mock_edge_chars
        else:
            raise ValueError
    mocker.patch("pydefect.cli.vasp.main_function.loadfn", side_effect=side_effect)

    mocker.patch("pydefect.cli.vasp.main_function.make_band_edge_state")
    mocker.patch("pydefect.cli.vasp.main_function.BandEdgeStates")


@pytest.mark.parametrize("skip_shallow", [False, True])
def test_make_defect_formation_energy(skip_shallow, tmpdir, mocker):
    tmpdir.chdir()
    mock_perfect_calc_results = mocker.Mock(spec=CalcResults)
    mock_perfect_calc_results.structure = Structure(Lattice.cubic(1), species=["H"] * 2, coords=[[0]*3] * 2)
    mock_perfect_calc_results.energy = 10
    mock_perfect_calc_results.vbm = 10
    mock_perfect_calc_results.cbm = 20

    mock_chem_pot_diag = mocker.patch("pydefect.cli.vasp.main_function.ChemPotDiag")
    mock_chem_pot_diag.from_yaml.return_value.abs_chem_pot_dict.return_value = {Element.H: 0}

    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_correction = mocker.Mock(spec=ExtendedFnvCorrection, autospec=True)

    def side_effect(key):
        if str(key) == "Va_O1_2/defect_entry.json":
            mock_defect_entry.name = "Va_H1"
            mock_defect_entry.charge = 2
            return mock_defect_entry
        elif str(key) == "Va_O1_2/calc_results.json":
            mock_calc_results.structure = Structure(Lattice.cubic(1), species=["H"], coords=[[0]*3])
            mock_calc_results.energy = 10
            return mock_calc_results
        elif str(key) == "Va_O1_2/correction.json":
            mock_correction.correction_energy = 20
            return mock_correction
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_function.loadfn", side_effect=side_effect)
    mock_band_edge_states = mocker.patch("pydefect.cli.vasp.main_function.BandEdgeStates")

    mock_unitcell = mocker.Mock(spec=Unitcell)
    mock_unitcell.vbm = 11
    mock_unitcell.cbm = 19

    args = Namespace(dirs=[Path("Va_O1_2")],
                     perfect_calc_results=mock_perfect_calc_results,
                     unitcell=mock_unitcell,
                     cpd_yaml="cpd.yaml",
                     label="A",
                     y_range=[-100, 100],
                     skip_shallow=skip_shallow,
                     supercell_edge=True,
                     label_line=True,
                     add_charges=True,
                     web_gui=False,
                     port=8050,
                     print=True)

    make_defect_formation_energy(args)

    if skip_shallow is True:
        mock_band_edge_states.from_yaml.assert_any_call(Path("Va_O1_2") / "band_edge_states.yaml")
    else:
        mock_loadfn.assert_any_call(Path("Va_O1_2") / "defect_entry.json")
        mock_loadfn.assert_any_call(Path("Va_O1_2") / "calc_results.json")
        mock_loadfn.assert_any_call(Path("Va_O1_2") / "correction.json")


def test_calc_defect_structure_info(mocker):

    mock_defect_entry = mocker.Mock(spec=DefectEntry)
    mock_defect_entry.structure = "a"
    mock_defect_entry.site_symmetry = "x"
    mock_calc_results = mocker.Mock(spec=CalcResults, autospec=True)
    mock_calc_results.structure = "b"
    mock_calc_results.site_symmetry = "y"

    def side_effect(key):

        if str(key) == "Va_O1_2/defect_entry.json":
            return mock_defect_entry
        elif str(key) == "Va_O1_2/calc_results.json":
            return mock_calc_results
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.vasp.main_function.loadfn",
                               side_effect=side_effect)
    mock_defect_str_info = mocker.patch(
        "pydefect.cli.vasp.main_function.make_defect_structure_info")

    mock_supercell_info = mocker.Mock(spec=SupercellInfo, autospec=True)
    mock_supercell_info.structure = "c"
    args = Namespace(supercell_info=mock_supercell_info,
                     dirs=[Path("Va_O1_2")],
                     symprec=None,
                     dist_tolerance=1.0)
    calc_defect_structure_info(args)
    mock_defect_str_info.assert_called_with("c", "a", "b",
                                            dist_tol=1.0,
                                            symprec=None,
                                            init_site_sym="x",
                                            final_site_sym="y")



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from copy import deepcopy

import numpy as np
import pytest
from pydefect.analyzer.defect_structure_comparator import SiteDiff
from pydefect.analyzer.defect_structure_info import \
    Displacement, DefectStructureInfo, fold_coords_in_structure, calc_drift, \
    make_defect_structure_info, symmetry_relation, SymmRelation, \
    unique_point_group, elem_indices_coords, DefectType, judge_defect_type, \
    calc_displacements
from pymatgen.core import Structure, Lattice
from vise.tests.helpers.assertion import assert_dataclass_almost_equal, \
    assert_msonable, assert_json_roundtrip
from vise.util.enum import ExtendedEnum
from vise.util.structure_symmetrizer import num_sym_op


@pytest.fixture
def structures():
    perfect = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Li", "U"],
        coords=[[0.25, 0.25, 0.25],
                [0.75, 0.75, 0.75],
                [0.5, 0.5, 0.5],
                [0, 0, 0]])

    initial = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.25, 0.25, 0.25],
                [0.75, 0.75, 0.75],
                [0.5, 0.5, 0.5],
                [0.0, 0.0, 0.0]])

    final = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.27, 0.25, 0.2501],
                [0.76, 0.75, 0.7501],
                [0.5, 0.5, 0.501],
                [0.0, 0.0, 0.0001]])
    return perfect, initial, final


@pytest.fixture
def displacements():
    return [None,
            Displacement(specie="He",
                         original_pos=(0.75, 0.75, 0.75),
                         final_pos=(0.76, 0.75, 0.75),
                         distance_from_defect=6.465437190360757,
                         disp_vector=(0.1, 0.0, 0.0),
                         displace_distance=0.1,
                         angle=124.9),
            Displacement(specie="Be",
                         original_pos=(0.5, 0.5, 0.5),
                         final_pos=(0.5, 0.5, 0.5009),
                         distance_from_defect=2.1355627039494767,
                         disp_vector=(0.0, 0.0, 0.009),
                         displace_distance=0.009,
                         angle=125.8),
            Displacement(specie="U",
                         original_pos=(0, 0, 0),
                         final_pos=(0.0, 0.0, 0.0),
                         distance_from_defect=6.525191802736529,
                         disp_vector=(0.0, 0.0, 0.0),
                         displace_distance=0.0,
                         angle=None)]


@pytest.fixture
def def_str_info(displacements):
    site_diff = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                         inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                         removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                         inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5009))])
    site_diff_from_init = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                                   inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                                   removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                                   inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5))])
    return DefectStructureInfo(initial_site_sym="3m",
                               final_site_sym="m",
                               site_diff=site_diff,
                               site_diff_from_initial=site_diff_from_init,
                               symprec=0.1,
                               dist_tol=0.2,
                               anchor_atom_idx=3,
                               drift_vector=(0.0, 0.0, 0.0001),
                               drift_dist=0.001,
                               center=(0.38, 0.375, 0.375175),
                               displacements=displacements)


def test_fold_coords():
    actual = Structure(Lattice.monoclinic(a=1, b=1, c=1, beta=20),
                       species=["H"],
                       coords=[[0.5, 0, 0.49]])
    fold_coords_in_structure(actual, [0, 0, 0])
    expected = Structure(Lattice.monoclinic(a=1, b=1, c=1, beta=20),
                         species=["H"],
                         coords=[[-0.5, 0, 0.49]])
    assert actual == expected


def test_calc_drift_dist(structures):
    perfect, _, final = structures
    anchor_atom_idx, drift_dist, vector_in_frac = \
        calc_drift(perfect, final, center=[0.4, 0.4, 0.4], d_to_p=[None, 1, 2, 3])
    assert anchor_atom_idx == 3
    assert drift_dist == 0.001
    np.testing.assert_almost_equal(vector_in_frac, np.array([0.0, 0.0, 0.0001]))


def test_calc_displacements():
    perf = Structure(Lattice.cubic(10), species=["H"], coords=[[0, 0, -0.01]])
    defect = Structure(Lattice.cubic(10), species=["H"], coords=[[0, 0, 10.98]])
    actual = calc_displacements(perf, defect, [0, 0, 0.98], d_to_p=[0])
    expected = [Displacement(specie="H",
                             original_pos=(0.0, 0.0, 0.99),
                             final_pos=(0.0, 0.0, 0.98),
                             distance_from_defect=0.1,
                             disp_vector=(0.0, 0.0, -0.1),
                             displace_distance=0.1,
                             angle=0.0)]
    assert_dataclass_almost_equal(actual[0], expected[0])


def test_elem_indices_coords(structures):
    actual = elem_indices_coords(structures[0], indices=[1, 2])
    expected = [("He", 1, (0.75, 0.75, 0.75)), ("Li", 2, (0.5, 0.5, 0.5))]
    assert actual == expected


def test_defect_type():
    assert_msonable(DefectType.vacancy)
    assert issubclass(DefectType, ExtendedEnum)


def test_judge_defect_type():
    """The structures remain as they could be used in the future"""
    perfect = Structure(Lattice.cubic(10), species=["H"]*4,
                        coords=[[0.0, 0.0, 0.0], [0.0, 0.5, 0.5],
                                [0.5, 0.0, 0.5], [0.5, 0.5, 0.0]])
    defect = Structure(Lattice.cubic(10), species=["H"]*3,
                       coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                               [0.5, 0.5, 0.0]])
    site_diff = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                  (2, "H", (0.5, 0.0, 0.5))],
                         inserted=[(1, "H", (0.25, 0.25, 0.5))],
                         removed_by_sub=[], inserted_by_sub=[])
    assert judge_defect_type(site_diff) == DefectType.vacancy_split

    defect_2 = Structure(Lattice.cubic(10), species=["H"]*5,
                         coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                                 [0.75, 0.25, 0.5], [0.25, 0.75, 0.5],
                                 [0.5, 0.5, 0.0]])
    site_diff_2 = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                    (2, "H", (0.5, 0.0, 0.5))],
                           inserted=[(1, "H", (0.25, 0.25, 0.5)),
                                     (2, "H", (0.75, 0.25, 0.5)),
                                     (3, "H", (0.25, 0.75, 0.5))],
                           removed_by_sub=[], inserted_by_sub=[])
    assert judge_defect_type(site_diff_2) == DefectType.interstitial_split

    defect_3 = Structure(Lattice.cubic(10), species=["H", "He", "H"],
                         coords=[[0.0, 0.0, 0.0], [0.25, 0.25, 0.5],
                                 [0.5, 0.5, 0.0]])
    site_diff_3 = SiteDiff(removed=[(1, "H", (0.0, 0.5, 0.5)),
                                    (2, "H", (0.5, 0.0, 0.5))],
                           inserted=[(1, "He", (0.25, 0.25, 0.5))],
                           removed_by_sub=[], inserted_by_sub=[])

    assert judge_defect_type(site_diff_3) == DefectType.unknown


def test_str_info_msonable(def_str_info):
    assert_msonable(def_str_info)


def test_calc_results_json_roundtrip(def_str_info, tmpdir):
    assert_json_roundtrip(def_str_info, tmpdir)


def test_symmetry_relation():
    for point_group in num_sym_op:
        pg = unique_point_group(point_group)
        assert symmetry_relation(pg, pg) == SymmRelation.same
    assert symmetry_relation("3m", "m") == SymmRelation.subgroup
    assert symmetry_relation("4", "2") == SymmRelation.subgroup
    assert symmetry_relation("-43m", "mm2") == SymmRelation.subgroup
    assert symmetry_relation("4/m", "2") == SymmRelation.subgroup


def test_make_defect_structure_info(structures, def_str_info):
    perfect, initial, final = structures
    actual = make_defect_structure_info(
         perfect, initial, final, dist_tol=0.2, symprec=0.1)
    assert_dataclass_almost_equal(actual, def_str_info)


def test_make_defect_structure_info_w_symms(structures, def_str_info):
    perfect, initial, final = structures
    actual = make_defect_structure_info(
        perfect, initial, final, dist_tol=0.2,
        init_site_sym="3m", final_site_sym="m")
    expected = deepcopy(def_str_info)
    expected.symprec = None
    assert_dataclass_almost_equal(actual, expected)


def test_make_def_str_info_symm_rel(structures, def_str_info):
    assert def_str_info.symm_relation == SymmRelation.subgroup


def test_make_def_str_info_same_config_from_init(structures, def_str_info):
    assert def_str_info.same_config_from_init is False


def test_def_str_info_defect_type(def_str_info):
    assert def_str_info.defect_type == DefectType.unknown


def test_repr(def_str_info):
    print(def_str_info)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import numpy as np
import pytest
from pydefect.input_maker.supercell_maker import SupercellMaker
from pydefect.util.error_classes import NotPrimitiveError
from pymatgen.core import Lattice, Structure


def test_create_supercell_tetragonal(tetra_close_to_cubic):
    cs = SupercellMaker(tetra_close_to_cubic)
    actual = cs.supercell.matrix
    expected =[[ 5,  5,  0],
               [-5,  5,  0],
               [ 0,  0,  5]]
    np.testing.assert_array_equal(actual, expected)


def test_create_supercell(a_centered_orthorhombic):
    cs = SupercellMaker(a_centered_orthorhombic)
    actual_lattice = cs.conv_structure.lattice
    expected = Lattice.orthorhombic(1, 4, 6)
    assert actual_lattice == expected

    actual = cs.supercell.matrix
    expected =[[6,  0,  0],
               [0,  2,  0],
               [0,  0,  1]]
    np.testing.assert_array_equal(actual, expected)


def test_create_supercell_raise_not_primitive_error(bcc):
    with pytest.raises(NotPrimitiveError):
        SupercellMaker(primitive_structure=bcc)


def test_create_supercell_matrix(a_centered_orthorhombic):
    matrix = [[2, 0, 0], [0, 1, 0], [0, 0, 1]]
    cs = SupercellMaker(a_centered_orthorhombic, matrix_to_conv_cell=matrix)
    actual = cs.supercell.matrix
    np.testing.assert_array_equal(actual, matrix)

    expected = [[2, 0, 0], [0, 1, -1], [0, 1, 1]]
    np.testing.assert_array_equal(cs.transformation_matrix, expected)


def test_create_supercell_generate_supercell_info(simple_cubic):
    cs = SupercellMaker(simple_cubic, matrix_to_conv_cell=[[2, 0, 0], [0, 2, 0], [0, 0, 2]])
    info = cs.supercell_info
    assert info.transformation_matrix == [[2, 0, 0], [0, 2, 0], [0, 0, 2]]
    assert info.sites["H1"].wyckoff_letter == "a"
    assert info.sites["H1"].equivalent_atoms == list(range(8))
    assert info.unitcell_structure == simple_cubic


def test_spglib_cyclic_behavior():
    structure = Structure.from_str("""Ca4 Sc2 Sb2 O12
1.0
5.469740 0.000000 0.000000
0.000000 5.632621 0.000000
-5.471566 0.000000 7.830251
O
12
direct
   0.242539  0.802107  0.449144
   0.242539  0.697893  0.949144
   0.757461  0.197893  0.550856
   0.757461  0.302107  0.050856
   0.859254  0.791326  0.05328
   0.859254  0.708674  0.55328
   0.140746  0.208674  0.94672
   0.140746  0.291326  0.44672
   0.656928  0.965329  0.256023
   0.656928  0.534671  0.756023
   0.343072  0.034671  0.743977
   0.343072  0.465329  0.243977
""", fmt="POSCAR")

    sm = SupercellMaker(structure)
    assert sm.conv_structure == structure


def test_transformation_matrix():
    structure = Structure.from_str("""
   1.0
    -2.5007160000000002    2.5007160000000002    3.6755999999999993
     2.5007160000000002   -2.5007159999999988    3.6755999999999993
     2.5007160000000002    2.5007159999999988   -3.6755999999999993
Si
   2    
Direct
  0.7500000000000001  0.2500000000000001  0.5000000000000000
  0.0000000000000000  0.0000000000000000  0.0000000000000000""", fmt="POSCAR")

    conv_structure = Structure.from_str("""
  1.0
5.001432 0.000000 0.000000
0.000000 5.001432 0.000000
0.000000 0.000000 7.351200
Si 
4 
direct
0.000000 0.500000 0.250000 Si
0.500000 1.000000 0.750000 Si
0.000000 0.000000 0.000000 Si
0.500000 0.500000 0.500000 Si""", fmt="POSCAR")

    matrix_to_conv_cell = [[3, 0, 0], [0, 3, 0], [0, 0, 2]]
    sm = SupercellMaker(structure, matrix_to_conv_cell=matrix_to_conv_cell)
    actual = structure * sm.transformation_matrix
    expected = conv_structure * matrix_to_conv_cell
    assert actual == expected





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from pydefect.corrections.manual_correction import NoCorrection, \
    ManualCorrection


def test_no_correction():
    assert NoCorrection().correction_energy == 0.0


def test_manual_correction():
    assert ManualCorrection(7.9).correction_energy == 7.9




# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
import fire
from monty.serialization import loadfn
from pydefect.analyzer.calc_results import CalcResults
from pydefect.analyzer.defect_structure_info import DefectStructureInfo, \
    fold_coords_in_structure
from pydefect.analyzer.vesta.create_vesta_file import VestaFile
from pydefect.input_maker.defect_entry import DefectEntry
from pydefect.input_maker.supercell_info import SupercellInfo
from pymatgen.core import Structure, DummySpecies, Element


def make_defect_vesta_file(defect_structure: Structure,
                           def_str_info: DefectStructureInfo,
                           title: str = None,
                           cutoff: float = 5.0,
                           show_disp_min: float = 0.1,
                           disp_enhance_factor: float = 3.0,
                           initial_filename: str = "defect_initial.vesta",
                           final_filename: str = "defect.vesta"):
    species, initial_coords, final_coords, site_info = [], [], [], []
    vectors, vector_colors = {}, []
    i = 1
    for f_idx, (s, disp) in enumerate(zip(defect_structure, def_str_info.displacements)):
        if disp and disp.distance_from_defect < cutoff:
            species.append(disp.specie)
            initial_coords.append(disp.original_pos)
            final_coords.append(s.frac_coords)
            info = f"{round(disp.distance_from_defect, 1)}"

            if disp and disp.displace_distance > show_disp_min:
                vectors[i] = (v * disp_enhance_factor for v in disp.disp_vector)
                # use 200 as max instead of 255 since (255, 255, 255) is white.
                rgb = int(200 * disp.angle / 180)
                vector_colors.append((rgb, rgb, 100))
                info += f"_{round(disp.displace_distance, 2)}"

            site_info.append(info)
            i += 1

    initial_local = Structure(defect_structure.lattice, species, initial_coords)
    final_local = Structure(defect_structure.lattice, species, final_coords)
    for s, i in zip(final_local, site_info):
        s.properties["name"] = i
    for _, specie, fcoord in def_str_info.site_diff.inserted:
        final_local.append(Element(specie), coords=fcoord)
        final_local[-1].properties["name"] = f"{specie}_i"

    for (_, i_specie, i_fcoord), (_, r_specie, r_fcoord) in zip(def_str_info.site_diff.inserted_by_sub, def_str_info.site_diff.removed_by_sub):
        initial_local.append(Element(r_specie), coords=r_fcoord)
        final_local.append(Element(i_specie), coords=i_fcoord)
        final_local[-1].properties["name"] = f"{i_specie}_{r_specie}"

    for _, specie, fcoord in def_str_info.site_diff.removed:
        initial_local.append(Element(specie), coords=fcoord)
        final_local.append(DummySpecies(), coords=fcoord)
        final_local[-1].properties["name"] = f"Va_{specie}"

    if def_str_info.site_diff.is_complex_defect:
        final_local.append(DummySpecies(), coords=def_str_info.center)
        final_local[-1].properties["name"] = f"center"

    fold_coords_in_structure(initial_local, def_str_info.center)
    fold_coords_in_structure(final_local, def_str_info.center)

    boundary = sum([[x - 0.5, x + 0.5] for x in def_str_info.center], [])

    VestaFile(initial_local, title, boundary=boundary).write_file(initial_filename)
    VestaFile(final_local, title, vectors=vectors, vector_colors=vector_colors, boundary=boundary).write_file(final_filename)


def make_defect_vesta_file_via_fire(supercell_info_json,
                                    defect_entry_json,
                                    calc_results_json,
                                    defect_str_info_json,
                                    cutoff: float = 5.0,
                                    show_disp_min: float = 0.1,
                                    disp_enhance_factor: float = 3.0,
                                    initial_filename: str = "defect_initial.vesta",
                                    final_filename: str = "defect.vesta"):
    supercell_info: SupercellInfo = loadfn(supercell_info_json)
    defect_entry: DefectEntry = loadfn(defect_entry_json)
    calc_results: CalcResults = loadfn(calc_results_json)
    defect_str_info = loadfn(defect_str_info_json)
    title = f"{defect_entry.full_name} in {supercell_info.structure.formula}"
    make_defect_vesta_file(calc_results.structure,
                           defect_str_info,
                           title,
                           cutoff,
                           show_disp_min,
                           disp_enhance_factor,
                           initial_filename,
                           final_filename)


if __name__ == '__main__':
    fire.Fire(make_defect_vesta_file_via_fire)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.analyzer.calc_results import CalcResults

from pydefect.analyzer.defect_energy import (
    DefectEnergy, CrossPoints, SingleDefectEnergy, make_defect_energies,
    defect_mpl_name, slide_energy, sanitize_defect_energies_for_plot,
    make_single_defect_energy, reservoir_energy, num_atom_differences,
    defect_plotly_name)
from pydefect.corrections.manual_correction import ManualCorrection
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import IStructure, Lattice, Element


@pytest.fixture
def defect_energy():
    return DefectEnergy(name="Va_O1",
                        charges=[0, 1, 2],
                        energies=[4, 2, -4],
                        corrections=[2, 1, 0])


def test_defect_energy_cross_points(defect_energy):
    actual = defect_energy.cross_points(1, 6)
    expected = CrossPoints([[5.0, 6.0]], [[1.0, -2.0], [6.0, 6.0]])
    assert actual == expected

    actual = defect_energy.cross_points(1, 6, 1)
    expected = CrossPoints([[4.0, 6.0]], [[0.0, -2.0], [5.0, 6.0]])
    assert actual == expected


def test_defect_energy_transition_levels(defect_energy):
    actual = defect_energy.transition_levels(base_e=0.0)
    expected = {(0, 1): 3.0, (0, 2): 5.0, (1, 2): 7.0}
    assert actual == expected

    actual = defect_energy.transition_levels(base_e=0.1)
    expected = {(0, 1): 2.9, (0, 2): 4.9, (1, 2): 6.9}
    assert actual == expected


def test_defect_energy_str(defect_energy):
    expected = """     Va_O1    0       4.0000       2.0000
     Va_O1    1       2.0000       1.0000
     Va_O1    2      -4.0000       0.0000"""
    assert str(defect_energy) == expected


def test_stable_charges(defect_energy):
    actual = defect_energy.stable_charges(ef_min=4.9, ef_max=5.1)
    assert actual == {0, 2}
    actual = defect_energy.stable_charges(ef_min=0.0, ef_max=0.1)
    assert actual == {2}


def test_pinning(defect_energy):
    assert defect_energy.pinning_level() == ((2.0, 2), (float("inf"), None))
    assert defect_energy.pinning_level(base_e=1.0) == ((1.0, 2), (float("inf"), None))


def test_energy_at_ef(defect_energy):
    assert defect_energy.energy_at_ef(ef=0.0) == (-4.0, 2)
    assert defect_energy.energy_at_ef(ef=10.0) == (6.0, 0)


@pytest.fixture
def cross_points():
    inner_cross_points = [[2, 30], [3, 40]]
    boundary_points = [[1, 10], [4, 40]]
    return CrossPoints(inner_cross_points, boundary_points)


def test_cross_points(cross_points):
    assert cross_points.all_sorted_points == [[1, 10], [2, 30], [3, 40], [4, 40]]
    assert cross_points.t_all_sorted_points == [[1, 2, 3, 4], [10, 30, 40, 40]]
    assert cross_points.t_inner_cross_points == [[2, 3], [30, 40]]
    assert cross_points.t_boundary_points == [[1, 4], [10, 40]]
    assert cross_points.charges == [20, 10, 0]
    assert cross_points.charge_list == [(None, 20), (20, 10), (10, 0), (0, None)]
    assert cross_points.annotated_charge_positions == \
           {20: [1.5, 20], 10: [2.5, 35], 0: [3.5, 40]}


def test_cross_points_str(cross_points):
    expected = """      1.0000      10.0000
      2.0000      30.0000
      3.0000      40.0000
      4.0000      40.0000"""
    assert str(cross_points) == expected


def test_generate_defect_energies():
    defect_energies = [
        SingleDefectEnergy("Va_O1", 0, -1, 1),
        SingleDefectEnergy("Va_Mg1", 0, -2, 2),
        SingleDefectEnergy("Va_O1", 1, 7, 3),
        SingleDefectEnergy("Va_Mg1", -1, -14, 4),
        SingleDefectEnergy("Va_O1", 2, 15, 5),
        SingleDefectEnergy("Va_Mg1", -2, -14, 6)]
    actual = make_defect_energies(defect_energies)
    expected = [DefectEnergy("Va_Mg1", [0, -1, -2], [-2, -14, -14], [2, 4, 6]),
                DefectEnergy("Va_O1", [0, 1, 2], [-1, 7, 15], [1, 3, 5])]
    assert actual == expected


def test_defect_mpl_name():
    assert defect_mpl_name(name="Va_O1") == "$V_{{\\rm O}1}$"
    assert defect_mpl_name(name="Mg_i1") == "${\\rm Mg}_{i1}$"


def test_defect_plotly_name():
    assert defect_plotly_name(name="Va_O1") == "<i>V</i><sub>O1</sub>"


def test_generate_defect_energies():
    energies = [DefectEnergy("Va_Mg1", [0], [0], [0]),
                DefectEnergy("Va_O1", [0], [0], [0]),
                DefectEnergy("Va_O2", [0], [0], [0]),
                DefectEnergy("Mg_i1", [0], [0], [0]),
                DefectEnergy("O_i1", [0], [0], [0])]
    actual = sanitize_defect_energies_for_plot(energies)
    expected = [DefectEnergy("$V_{{\\rm Mg}}$", [0], [0], [0]),
                DefectEnergy("$V_{{\\rm O}1}$", [0], [0], [0]),
                DefectEnergy("$V_{{\\rm O}2}$", [0], [0], [0]),
                DefectEnergy("${\\rm Mg}_{i}$", [0], [0], [0]),
                DefectEnergy("${\\rm O}_{i}$", [0], [0], [0])]
    assert actual == expected

    energies = [DefectEnergy("Mg_i1", [0], [0], [0]),
                DefectEnergy("Mg_i2", [0], [0], [0]),]
    actual = sanitize_defect_energies_for_plot(energies)
    expected = [DefectEnergy("${\\rm Mg}_{i1}$", [0], [0], [0]),
                DefectEnergy("${\\rm Mg}_{i2}$", [0], [0], [0])]
    assert actual == expected

    energies = [DefectEnergy("Mg_i1", [0], [0], [0]),
                DefectEnergy("Mg_i2", [0], [0], [0]),]
    actual = sanitize_defect_energies_for_plot(energies, for_plotly=True)
    expected = [DefectEnergy("Mg<sub>i1</sub>", [0], [0], [0]),
                DefectEnergy("Mg<sub>i2</sub>", [0], [0], [0])]
    assert actual == expected


def test_slide_energy():
    energies = [DefectEnergy("Va_Mg1", [0], [0], [0]),
                DefectEnergy("Va_O1", [1], [0], [0]),
                DefectEnergy("Va_O2", [2], [0], [0]),
                DefectEnergy("Mg_i1", [0], [0], [0])]
    actual = slide_energy(energies, 1.0)
    expected = [DefectEnergy("Va_Mg1", [0], [0], [0]),
                DefectEnergy("Va_O1", [1], [1.0], [0]),
                DefectEnergy("Va_O2", [2], [2.0], [0]),
                DefectEnergy("Mg_i1", [0], [0], [0])]
    assert actual == expected

    """
    TODO
    - Evaluate the crossing points at given Fermi level range.
    """


def test_make_single_defect_energy(mocker):
    perfect = mocker.Mock(CalcResults, autospec=True)
    perfect.energy = 1.0
    perfect.structure = IStructure(Lattice.cubic(1), ["H"] * 2, [[0] * 3] * 2)

    defect = mocker.Mock(CalcResults, autospec=True)
    defect.energy = 3.0
    defect.structure = IStructure(Lattice.cubic(1), ["H"], [[0] * 3])

    name, charge = "Va_H1", 5
    defect_entry = mocker.Mock(DefectEntry, autospec=True)
    defect_entry.name = name
    defect_entry.charge = charge

    single_defect_energy = \
        make_single_defect_energy(perfect=perfect,
                                  defect=defect,
                                  defect_entry=defect_entry,
                                  abs_chem_pot={Element.H: 100},
                                  correction=ManualCorrection(7.0))
    energy = 3.0 - 1.0 + 100

    assert isinstance(single_defect_energy, SingleDefectEnergy)
    assert single_defect_energy == SingleDefectEnergy(name, charge, energy, 7.0)


def test_reservoir_energy():
    num_atom_diff = {Element.He: 1, Element.Li: -1}
    abs_chem_pot = {Element.He: 10.0, Element.Li: 100.0, Element.H: 1000.0}
    actual = reservoir_energy(num_atom_diff, abs_chem_pot)
    assert actual == -10.0 + 100.0


def test_num_atom_diff():
    s1 = IStructure(Lattice.cubic(1), ["H", "He"], [[0] * 3] * 2)
    s2 = IStructure(Lattice.cubic(1), ["H", "Li"], [[0] * 3] * 2)
    assert num_atom_differences(s1, s2) == {Element.He: 1, Element.Li: -1}


# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np
import pytest
from pydefect.cli.vasp.make_band_edge_eigenvalues import \
    make_band_edge_eigenvalues
from pymatgen.electronic_structure.core import Spin
from pymatgen.io.vasp import Vasprun


@pytest.fixture
def band_edge_eigenvalues(mocker):
    mock_vasprun = mocker.Mock(spec=Vasprun, autospec=True)
    mock_vasprun.actual_kpoints = [[0.0, 0.0, 0.0]]
    mock_defaults = mocker.patch("pydefect.cli.vasp.make_band_edge_eigenvalues.defaults")
    mock_defaults.eigval_range = 3.0
    mock_vasprun.eigenvalues = {Spin.up: np.array([[[-3.01, 1.],
                                           [-2.99, 1.],
                                           [ 8.01, 0.],
                                           [10.00, 0.]],
                                          [[-3.01, 1.],
                                           [-2.99, 1.],
                                           [ 7.99, 0.],
                                           [10.00, 0.]]]),
                                Spin.down: np.array([[[-3.01, 1.],
                                           [-2.99, 1.],
                                           [ 7.99, 0.],  # only change here from Spin.up
                                           [10.00, 0.]],
                                          [[-3.01, 1.],
                                           [-2.99, 1.],
                                           [ 7.99, 0.],
                                           [10.00, 0.]]])}
    return make_band_edge_eigenvalues(mock_vasprun, vbm=0.0, cbm=5.0)


def test_make_band_edge_eigenvalues_kpt_coords(band_edge_eigenvalues):
    assert band_edge_eigenvalues.kpt_coords == [(0.0, 0.0, 0.0)]


def test_make_band_edge_eigenvalues_eigenvalues(band_edge_eigenvalues):
    actual = band_edge_eigenvalues.energies_and_occupations
    expected = [[[[-2.99, 1.], [8.01, 0.]], [[-2.99, 1.], [7.99, 0.]]],
                [[[-2.99, 1.], [7.99, 0.]], [[-2.99, 1.], [7.99, 0.]]]]
    np.testing.assert_array_almost_equal(actual, expected)


def test_make_band_edge_eigenvalues_lowest_band_index(band_edge_eigenvalues):
    assert band_edge_eigenvalues.lowest_band_index == 1



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from monty.serialization import loadfn
from pydefect.analyzer.band_edge_states import EdgeCharacter, EdgeState
from pydefect.analyzer.make_band_edge_state import make_band_edge_state, \
    are_orbitals_similar


@pytest.fixture
def no_in_gap_character():
    return EdgeCharacter(vbm_participation_ratio=0.1,
                         cbm_participation_ratio=0.2,
                         hob_bottom_e=-0.1,
                         lub_top_e=1.1,
                         vbm=0.0,
                         cbm=1.0,
                         vbm_orbitals={"Mn": [0.5, 0.6]},
                         cbm_orbitals={"O": [0.7, 0.8]})


@pytest.fixture
def donor_phs_character():
    return EdgeCharacter(vbm_participation_ratio=0.1,
                         cbm_participation_ratio=0.2,
                         hob_bottom_e=0.9,
                         lub_top_e=1.1,
                         vbm=None,
                         cbm=None,
                         vbm_orbitals={"O": [0.5, 0.6]},
                         cbm_orbitals={"O": [0.7, 0.8]})


@pytest.fixture
def acceptor_phs_character():
    return EdgeCharacter(vbm_participation_ratio=0.1,
                         cbm_participation_ratio=0.2,
                         hob_bottom_e=-0.2,
                         lub_top_e=0.1,
                         vbm=None,
                         cbm=None,
                         vbm_orbitals={"Mn": [0.5, 0.6]},
                         cbm_orbitals={"Mn": [0.7, 0.8]})


@pytest.fixture
def in_gap_state_character():
    return EdgeCharacter(vbm_participation_ratio=0.5,
                         cbm_participation_ratio=0.5,
                         hob_bottom_e=0.4,
                         lub_top_e=0.6,
                         vbm=0.45,
                         cbm=0.55,
                         vbm_orbitals={"Mn": [0.3, 0.2], "O": [0.3, 0.1]},
                         cbm_orbitals={"Mn": [0.3, 0.2], "O": [0.4, 0.2]})


# @pytest.fixture
# def unknown_state_character():
#     return EdgeCharacter(vbm_participation_ratio=0.1,
#                          cbm_participation_ratio=0.2,
#                          hob_bottom_e=0.4,
#                          lub_top_e=0.6,
#                          vbm=0.45,
#                          cbm=0.55,
#                          vbm_orbitals={"Mn": [0.3, 0.2], "O": [0.3, 0.1]},
#                          cbm_orbitals={"Mn": [0.3, 0.2], "O": [0.4, 0.2]})


@pytest.fixture
def perfect_character():
    return EdgeCharacter(hob_bottom_e=-0.1,
                         lub_top_e=1.1,
                         vbm=0.0,
                         cbm=1.0,
                         vbm_orbitals={"Mn": [0.5, 0.6]},
                         cbm_orbitals={"O": [0.7, 0.8]})


def test_make_band_edge_state_no_in_gap(no_in_gap_character, perfect_character):
    actual = make_band_edge_state(no_in_gap_character, perfect_character)
    assert actual == EdgeState.no_in_gap


def test_make_band_edge_state_donor_phs(donor_phs_character, perfect_character):
    actual = make_band_edge_state(donor_phs_character, perfect_character)
    assert actual == EdgeState.donor_phs


def test_make_band_edge_state_acceptor_phs(acceptor_phs_character,
                                           perfect_character):
    actual = make_band_edge_state(acceptor_phs_character, perfect_character)
    assert actual == EdgeState.acceptor_phs


def test_make_band_edge_state_in_gap_state(in_gap_state_character,
                                           perfect_character):
    actual = make_band_edge_state(in_gap_state_character, perfect_character)
    assert actual == EdgeState.in_gap_state


# def test_make_band_edge_state_unknown_state(unknown_state_character, perfect_character):
#     actual = make_band_edge_state(unknown_state_character, perfect_character)
#     assert actual == EdgeState.unknown


def test_are_orbitals_similar(mocker):
    mock = mocker.patch("pydefect.analyzer.make_band_edge_state.defaults")
    mock.similar_orb_criterion = 27.0  # = 1 + 2 + 6 + 7 + 8 + 1 + 1 + 1
    orb_1 = {"Mn": [0.0, 1.0, 2.0], "O": [3.0, 4.0, 5.0]}
    orb_2 = {"Mg": [-6.0, -7.0, -8.0], "O": [4.0, 3.0, 6.0]}
    assert are_orbitals_similar(orb_1, orb_2) is False

    mock.similar_orb_criterion = 27.1
    assert are_orbitals_similar(orb_1, orb_2) is True


dir_path = Path(__file__).parent / "band_edge_database" / "NaCl"


def test_edge_character_nacl_acceptor_phs():
    ref = loadfn(dir_path / "perfect_edge_characters.json")[0]  # only spin up
    target = loadfn(dir_path / "Va_Na1_1_edge_characters.json")[0]
    assert make_band_edge_state(target, ref) == EdgeState.acceptor_phs


def test_edge_character_nacl_no_in_gap():
    ref = loadfn(dir_path / "perfect_edge_characters.json")[0]
    target = loadfn(dir_path / "Va_Na1_-1_edge_characters.json")[0]
    assert make_band_edge_state(target, ref) == EdgeState.no_in_gap





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
import math
import warnings
from dataclasses import dataclass
from typing import Optional, List, Tuple

import numpy as np
from monty.json import MSONable
from numpy.linalg import inv
from pydefect.analyzer.defect_structure_comparator import \
    DefectStructureComparator, SiteDiff
from pymatgen.core import Structure, Lattice, PeriodicSite
from pymatgen.symmetry.groups import SpaceGroup
from tabulate import tabulate
from vise.util.enum import ExtendedEnum
from vise.util.logger import get_logger
from vise.util.mix_in import ToJsonFileMixIn
from vise.util.structure_symmetrizer import StructureSymmetrizer

logger = get_logger(__name__)


def folded_coords(site: PeriodicSite, center: List[float]):
    _, image = site.distance_and_image_from_frac_coords(center)
    return tuple(site.frac_coords - image)


def fold_coords_in_structure(structure: Structure, center: List[float]):
    for site in structure:
        _, image = site.distance_and_image_from_frac_coords(center)
        site.frac_coords -= image


def calc_drift(perfect: Structure,
               defect: Structure,
               center: List[float],
               d_to_p: List[int]):
    distances = []
    for site in defect:
        distances.append(site.distance_and_image_from_frac_coords(center)[0])
    anchor_atom_idx = int(np.argmax(distances))
    p_anchor_atom_idx = d_to_p[anchor_atom_idx]
    distance, image = defect[anchor_atom_idx].distance_and_image_from_frac_coords(perfect[p_anchor_atom_idx].frac_coords)
    drift_vector = (defect[anchor_atom_idx].frac_coords
                    - perfect[p_anchor_atom_idx].frac_coords - image)
    return anchor_atom_idx, distance, drift_vector


def calc_displacements(perfect: Structure,
                       defect: Structure,
                       center: List[float],
                       d_to_p: List[int]):
    result = []
    lattice: Lattice = defect.lattice
    for d, p in enumerate(d_to_p):
        if p is None:
            result.append(None)
        else:
            p_elem = str(perfect[p].specie)
            d_elem = str(defect[d].specie)
            if p_elem != d_elem:
                result.append(None)
                continue
            initial_pos = folded_coords(perfect[p], center)
            final_pos = folded_coords(defect[d], initial_pos)

            initial_pos_vec = lattice.get_cartesian_coords(np.array(initial_pos) - center)
            disp_dist, t = lattice.get_distance_and_image(
                defect.frac_coords[d], perfect.frac_coords[p])
            disp_vec = lattice.get_cartesian_coords(
                defect.frac_coords[d] - perfect.frac_coords[p] - t)

            inner_prod = sum(initial_pos_vec * disp_vec)
            ini_dist = np.linalg.norm(initial_pos_vec)
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                # ignore "RuntimeWarning: invalid value encountered in double_scalars"
                cos = round(inner_prod / (ini_dist * disp_dist), 10)
            angle = float(round(180 * (1 - np.arccos(cos) / np.pi), 1))
            if math.isnan(angle):
                angle = None

            result.append(Displacement(specie=p_elem,
                                       original_pos=initial_pos,
                                       final_pos=final_pos,
                                       distance_from_defect=ini_dist,
                                       disp_vector=tuple(disp_vec),
                                       displace_distance=disp_dist,
                                       angle=angle))
    return result


def remove_dot(x):
    return "".join([s for s in x if s != "."])


def elem_indices_coords(structure: Structure,
                        indices: List[int]
                        ) -> List[Tuple[str, int, Tuple[float, float, float]]]:
    result = []
    for i in indices:
        site = structure[i]
        result.append((str(site.specie), i, tuple(site.frac_coords)))
    return result


class DefectType(MSONable, ExtendedEnum):
    vacancy = "vacancy"
    interstitial = "interstitial"
    substituted = "substituted"
    vacancy_split = "vacancy_split"
    interstitial_split = "interstitial_split"
    unknown = "unknown"


def judge_defect_type(site_diff: SiteDiff):
    if site_diff.is_vacancy:
        return DefectType.vacancy
    elif site_diff.is_interstitial:
        return DefectType.interstitial
    elif site_diff.is_substituted:
        return DefectType.substituted

    elements_involved = set()
    for _, elem, _ in site_diff.removed + site_diff.inserted:
        elements_involved.add(elem)

    if len(elements_involved) == 1 and not site_diff.removed_by_sub:
        if len(site_diff.removed) - len(site_diff.inserted) == 1:
            return DefectType.vacancy_split
        elif len(site_diff.removed) - len(site_diff.inserted) == -1:
            return DefectType.interstitial_split

    return DefectType.unknown


def make_defect_structure_info(perfect: Structure,
                               initial: Structure,
                               final: Structure,
                               dist_tol: float,
                               symprec: float = None,
                               init_site_sym: str = None,
                               final_site_sym: str = None
                               ) -> "DefectStructureInfo":
    final = final.copy()
    if symprec:
        initial_symmetrizer = StructureSymmetrizer(initial, symprec)
        final_symmetrizer = StructureSymmetrizer(final, symprec)
        init_site_sym = initial_symmetrizer.point_group
        final_site_sym = final_symmetrizer.point_group
    else:
        if not (init_site_sym and final_site_sym):
            raise AssertionError("Need to set both initial and final site "
                                 "symmetry, when symprec is not set.")

    assert perfect.lattice == initial.lattice == final.lattice
    comp_w_perf = DefectStructureComparator(final, perfect, dist_tol)
    comp_w_init = DefectStructureComparator(final, initial, dist_tol)
    d_to_p = comp_w_init.d_to_p

    center = comp_w_perf.defect_center_coord
    anchor_idx, drift_d, drift_vec = calc_drift(initial, final, center, d_to_p)
    center -= drift_vec
    for site in final:
        site.frac_coords -= drift_vec

    displacements = calc_displacements(initial, final, center, d_to_p)

    return DefectStructureInfo(
        initial_site_sym=unique_point_group(remove_dot(init_site_sym)),
        final_site_sym=unique_point_group(final_site_sym),
        site_diff=comp_w_perf.make_site_diff(),
        site_diff_from_initial=comp_w_init.make_site_diff(),
        symprec=symprec,
        dist_tol=dist_tol,
        anchor_atom_idx=anchor_idx,
        drift_vector=tuple(drift_vec),
        drift_dist=drift_d,
        center=tuple(center),
        displacements=displacements)


class SymmRelation(MSONable, ExtendedEnum):
    same = "same"
    subgroup = "subgroup"
    supergroup = "supergroup"
    another = "another"


@dataclass
class Displacement(MSONable):
    specie: str
    original_pos: Tuple[float, float, float]
    final_pos: Tuple[float, float, float]
    distance_from_defect: float
    disp_vector: Tuple[float, float, float]
    displace_distance: float
    angle: Optional[float]


def unique_point_group(pg):
    if pg == "2mm" or pg == "m2m":
        return "mm2"
    if pg == "-4m2":
        return "-42m"
    if pg == "m3":
        return "m-3"
    return pg


def symmetry_relation(initial_point_group, final_point_group):
    """ Check the point group symmetry relation using the space group relation
    implemented in pymatgen.
    """
    if initial_point_group in ["3m", "-3m"]:
        initial_point_group += "1"
    if final_point_group in ["3m", "-3m"]:
        final_point_group += "1"

    initial = SpaceGroup(f"P{initial_point_group}")
    final = SpaceGroup(f"P{final_point_group}")
    if initial == final:
        return SymmRelation.same
    elif final.is_subgroup(initial):
        return SymmRelation.subgroup
    elif final.is_supergroup(initial):
        return SymmRelation.supergroup
    else:
        return SymmRelation.another


@dataclass
class DefectStructureInfo(MSONable, ToJsonFileMixIn):
    initial_site_sym: str
    final_site_sym: str
    site_diff: SiteDiff
    site_diff_from_initial: SiteDiff
    symprec: float
    dist_tol: float
    anchor_atom_idx: int
    drift_vector: Tuple[float, float, float]
    drift_dist: float
    center: Tuple[float, float, float]
    displacements: List[Displacement]

    @property
    def symm_relation(self):
        return symmetry_relation(self.initial_site_sym, self.final_site_sym)

    @property
    def same_config_from_init(self):
        return self.site_diff_from_initial.is_no_diff

    @property
    def defect_type(self):
        return judge_defect_type(self.site_diff)

    def __repr__(self):
        lines = ["",
                 f"Defect type: {self.defect_type}",
                 f"Site symmetry: {self.initial_site_sym} "
                 f"-> {self.final_site_sym} ({self.symm_relation})",
                 f"Is same configuration: {self.same_config_from_init}",
                 f"Drift distance: {self.drift_dist:5.3f}"]
        center_coords = [round(c, 3) for c in self.center]
        lines.append(f"Defect center: {center_coords}")

        lines.append("Removed atoms:")
        x = []
        for i, elem, coords in self.site_diff.removed:
            x.append([i, elem] + [round(fc, 2) for fc in coords])
        lines.append(tabulate(x))

        lines.append("Added atoms:")
        x = []
        for i, elem, coords in self.site_diff.inserted:
            x.append([i, elem] + [round(fc, 2) for fc in coords])
        lines.append(tabulate(x))

        if self.same_config_from_init is False:
            lines.append("Initially removed atoms:")
            x = []
            for i, elem, coords in self.site_diff_from_initial.removed:
                x.append([i, elem] + [round(fc, 2) for fc in coords])
            lines.append(tabulate(x))

            lines.append("Initially added atoms:")
            x = []
            for i, elem, coords in self.site_diff_from_initial.inserted:
                x.append([i, elem] + [round(fc, 2) for fc in coords])
            lines.append(tabulate(x))

        lines.append("Displacements")
        idxs = [[i, d] for i, d in enumerate(self.displacements) if d is not None]
        x = []
        for final_idx, d in sorted(idxs, key=lambda y: y[1].distance_from_defect):
            i_pos = [round(c, 3) for c in d.original_pos]
            f_pos = [round(c, 3) for c in d.final_pos]
            angle = int(round(d.angle, -1)) if d.angle else ""
            x.append([d.specie, round(d.distance_from_defect, 2),
                      round(d.displace_distance, 2), angle, final_idx] + i_pos
                     + ["->"] + f_pos)
        lines.append(tabulate(x))

        return "\n".join(lines)





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from itertools import cycle
from pathlib import Path

from monty.design_patterns import singleton
from vise.user_settings import UserSettings


@singleton
class Defaults:
    def __init__(self):
        self._symmetry_length_tolerance = 0.1
        self._symmetry_angle_tolerance = 5.0
        self._ewald_accuracy = 15.0
        self._e_above_hull = 1e-5
        self._cutoff_distance_factor = 1.3
        self._displace_distance = 0.2
        self._dist_tol = 1.0
        self._ele_neg_diff = 2.0
        self._kpoint_density = 5.0
        self._band_gap_criterion = 0.2  # in eV
        self._similar_orb_criterion = 0.12
        self._similar_energy_criterion = 0.2  # in eV
        self._eigval_range = 1.0
        self._integer_criterion = 0.1
        self._defect_energy_colors = \
            ["xkcd:blue", "xkcd:brown", "xkcd:crimson", "xkcd:darkgreen",
             "xkcd:gold", "xkcd:magenta", "xkcd:orange", "xkcd:darkblue",
             "xkcd:navy", "xkcd:red", "xkcd:olive", "xkcd:black", "xkcd:indigo"]
        self._outcar = "OUTCAR"
        self._contcar = "CONTCAR"
        self._vasprun = "vasprun.xml"
        self._procar = "PROCAR"

        user_settings = UserSettings(yaml_filename="pydefect.yaml")
        self.yaml_files = user_settings.yaml_files_from_root_dir
        self.user_settings = user_settings.user_settings

        for k, v in self.user_settings.items():
            if hasattr(self, k):
                self.__setattr__("_" + k, v)

    @property
    def symmetry_length_tolerance(self):
        return self._symmetry_length_tolerance

    @property
    def symmetry_angle_tolerance(self):
        return self._symmetry_angle_tolerance

    @property
    def ewald_accuracy(self):
        return self._ewald_accuracy

    @property
    def e_above_hull(self):
        return self._e_above_hull

    @property
    def cutoff_distance_factor(self):
        return self._cutoff_distance_factor

    @property
    def displace_distance(self):
        return self._displace_distance

    @property
    def dist_tol(self):
        return self._dist_tol

    @property
    def ele_neg_diff(self):
        return self._ele_neg_diff

    @property
    def kpoint_density(self):
        return self._kpoint_density

    @property
    def similar_orb_criterion(self):
        return self._similar_orb_criterion

    @property
    def similar_energy_criterion(self):
        return self._similar_energy_criterion

    @property
    def eigval_range(self):
        return self._eigval_range

    @property
    def band_gap_criterion(self):
        return self._band_gap_criterion

    @property
    def integer_criterion(self):
        return self._integer_criterion

    @property
    def defect_energy_colors(self):
        return cycle(self._defect_energy_colors)

    @property
    def outcar(self):
        return Path(self._outcar)

    @property
    def contcar(self):
        return Path(self._contcar)

    @property
    def vasprun(self):
        return Path(self._vasprun)

    @property
    def procar(self):
        return Path(self._procar)


defaults = Defaults()



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import pytest
from pydefect.analyzer.defect_energy import DefectEnergy
from pydefect.analyzer.defect_energy_plotter import DefectEnergyMplPlotter, \
    DefectEnergiesMplSettings, DefectEnergyPlotlyPlotter
from pydefect.defaults import defaults
from pymatgen.util.string import latexify
from vise.util.dash_helper import show_png


try:
    import psutil
    PSUTIL_NOT_PRESENT = False
except ModuleNotFoundError:
    PSUTIL_NOT_PRESENT = True


def test_defect_energies_mpl_settings():
    mpl_defaults = DefectEnergiesMplSettings()
    assert next(mpl_defaults.colors) == next(defaults.defect_energy_colors)

    mpl_defaults = DefectEnergiesMplSettings(colors=["black"])
    assert next(mpl_defaults.colors) == "black"


@pytest.fixture
def defect_energy_plotters():
    va_o = DefectEnergy(name="Va_O1", charges=[0, 1, 2], energies=[5, 2, -5], corrections=[1, 1, 1])
    va_mg = DefectEnergy(name="Va_Mg1", charges=[-2, -1, 0], energies=[5, 2, 0], corrections=[-1, -1, -1])
    mg_i = DefectEnergy(name="Mg_i1", charges=[1], energies=[4], corrections=[1])

    d = dict(title=latexify("MgAl2O4"),
             defect_energies=[va_o, va_mg, mg_i],
             vbm=1.5, cbm=5.5, supercell_vbm=1.0, supercell_cbm=6.0,
             supercell_edge=True)

    return DefectEnergyMplPlotter(**d), DefectEnergyPlotlyPlotter(**d)


def test_defect_energies_mpl_actual_plot(defect_energy_plotters):
    plotter, _ = defect_energy_plotters
    plotter.construct_plot()
    plotter.plt.show()


@pytest.mark.skipif(PSUTIL_NOT_PRESENT, reason="psutil does not exist")
def test_defect_energies_plotly_actual_plot(defect_energy_plotters):
    _, plotter = defect_energy_plotters
    fig = plotter.create_figure()
    show_png(fig)


"""
TODO
- Show all energies with thinner lines.
- show_transition_levels

- Add labels with optimize positions
https://github.com/cphyc/matplotlib-label-lines
https://github.com/Phlya/adjustText

- Add boundary points with open circles for shallow defects.
DONE
"""


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import numpy as np
from pydefect.analyzer.defect_charge_distribution import RadialDist
from pymatgen.electronic_structure.core import Spin
from pymatgen.core import Structure, Lattice
from pymatgen.io.vasp import Chgcar


def test():
    structure = Structure(Lattice.cubic(3), ["H"], [[0, 0, 0]])
    data = {"total": np.array([[[3]*3]*3]*3), "diff": np.array([[[-1]*3]*3]*3)}
    chgcar = Chgcar(structure, data)

    rad = RadialDist(chgcar, [0, 0, 0])
    assert len(rad.distances_data) == 1 + 6 + 12
    np.testing.assert_almost_equal(rad.distances_data[0][0], np.array([-1/3, -1./3, 0.0]))
    assert rad.distances_data[0][1] == np.sqrt(2)
    assert rad.distances_data[0][2] == 24
    np.testing.assert_almost_equal(rad.distances_data[0][3], np.array([-1, -1, 0.0]))

    hist_data, half_point, summed = rad.histogram(Spin.up)
    np.testing.assert_almost_equal(hist_data[0], [5.00000000e-02, 1.16355283e-03])





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.
from pathlib import Path

import pytest
from pydefect.analyzer.band_edge_states import BandEdgeEigenvalues, \
    BandEdgeStates, EdgeState, EdgeCharacter, EdgeCharacters
from vise.tests.helpers.assertion import assert_msonable, assert_json_roundtrip


@pytest.fixture
def band_edge_eigenvalues():
    return BandEdgeEigenvalues(energies_and_occupations=[[[[0.0, 1.0]]]],
                               kpt_coords=[(0.0, 0.0, 0.0)],
                               lowest_band_index=10)


def test_band_edge_eigenvalues_msonable(band_edge_eigenvalues):
    assert_msonable(band_edge_eigenvalues)


def test_band_edge_eigenvalues_to_json_file(band_edge_eigenvalues, tmpdir):
    assert_json_roundtrip(band_edge_eigenvalues, tmpdir)


@pytest.fixture
def band_edge_states():
    return BandEdgeStates([EdgeState.donor_phs, EdgeState.no_in_gap])


def test_band_edge_states_is_shallow(band_edge_states):
    assert band_edge_states.is_shallow is True
    actual = BandEdgeStates([EdgeState.in_gap_state, EdgeState.no_in_gap])
    assert actual.is_shallow is False


def test_band_edge_states_msonable(band_edge_states):
    assert_msonable(band_edge_states)


def test_band_edge_states_to_json_file(band_edge_states, tmpdir):
    assert_json_roundtrip(band_edge_states, tmpdir)


def test_band_edge_states_yaml(band_edge_states, tmpdir):
    print(tmpdir)
    tmpdir.chdir()
    band_edge_states.to_yaml()
    expected = """up: Donor PHS
down: No in-gap state"""
    assert Path("band_edge_states.yaml").read_text() == expected

    actual = band_edge_states.from_yaml("band_edge_states.yaml")
    assert actual == band_edge_states


@pytest.fixture
def edge_characters():
    edge_character = EdgeCharacter(vbm_participation_ratio=0.1,
                                   cbm_participation_ratio=0.2,
                                   hob_bottom_e=0.3,
                                   lub_top_e=0.4,
                                   vbm=0.5,
                                   cbm=0.6,
                                   vbm_orbitals={"Mn": [0.5, 0.6]},
                                   cbm_orbitals={"Mn": [0.7, 0.8]})
    return EdgeCharacters([edge_character])


def test_edge_characters_msonable(edge_characters):
    assert_msonable(edge_characters)


def test_edge_characters_to_json_file(edge_characters, tmpdir):
    assert_json_roundtrip(edge_characters, tmpdir)


def test_str(edge_characters):
    assert str(edge_characters) == """  up: EdgeCharacter(hob_bottom_e=0.3, lub_top_e=0.4, vbm=0.5, cbm=0.6, vbm_orbitals={'Mn': [0.5, 0.6]}, cbm_orbitals={'Mn': [0.7, 0.8]}, vbm_participation_ratio=0.1, cbm_participation_ratio=0.2)"""





# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

from pydefect.analyzer.defect_structure_comparator import SiteDiff
from pydefect.analyzer.defect_structure_info import DefectStructureInfo, \
    Displacement
from pydefect.cli.vasp.util_commands.create_defect_vesta_file import \
    make_defect_vesta_file
from pymatgen.core import Structure, Lattice


def test_defect_vesta_file(tmpdir):
    print(tmpdir)
    site_diff = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                         inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                         removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                         inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5009))])
    site_diff_from_init = SiteDiff(removed=[(0, 'H', (0.25, 0.25, 0.25))],
                                   inserted=[(0, 'H', (0.27, 0.25, 0.25))],
                                   removed_by_sub=[(2, 'Li', (0.5, 0.5, 0.5))],
                                   inserted_by_sub=[(2, 'Be', (0.5, 0.5, 0.5))])
    displacements = [None,
            Displacement(specie="He",
                         original_pos=(0.75, 0.75, 0.75),
                         final_pos=(0.76, 0.75, 0.75),
                         distance_from_defect=6.465437190360757,
                         disp_vector=(0.1, 0.0, 0.0),
                         displace_distance=0.09999999999999964,
                         angle=124.9),
            Displacement(specie="Be",
                         original_pos=(0.5, 0.5, 0.5),
                         final_pos=(0.5, 0.5, 0.5009),
                         distance_from_defect=2.1355627039494767,
                         disp_vector=(0.0, 0.0, 0.009),
                         displace_distance=0.009000000000000341,
                         angle=125.8),
            Displacement(specie="U",
                         original_pos=(0, 0, 0),
                         final_pos=(0.0, 0.0, 0.0),
                         distance_from_defect=6.525191802736529,
                         disp_vector=(0.0, 0.0, 0.0),
                         displace_distance=0.0,
                         angle=None)]

    def_str_info = DefectStructureInfo(initial_site_sym="3m",
                               final_site_sym="m",
                               site_diff=site_diff,
                               site_diff_from_initial=site_diff_from_init,
                               symprec=0.1,
                               dist_tol=0.2,
                               anchor_atom_idx=3,
                               drift_vector=(0.0, 0.0, 0.0001),
                               drift_dist=0.001,
                               center=(0.38, 0.375, 0.37517500000000004),
                               displacements=displacements)
    structure = Structure(
        Lattice.cubic(10),
        species=["H", "He", "Be", "U"],
        coords=[[0.27, 0.25, 0.2501],
                [0.76, 0.75, 0.7501],
                [0.5, 0.5, 0.501],
                [0.0, 0.0, 0.0001]])
    tmpdir.chdir()
    make_defect_vesta_file(defect_structure=structure, def_str_info=def_str_info, disp_enhance_factor=100)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import List

import numpy as np
from numpy.linalg import inv
from pydefect.input_maker.supercell_info import SupercellInfo, Interstitial
from pydefect.util.error_classes import NotPrimitiveError
from pymatgen.core import Structure, Element
from vise.util.structure_symmetrizer import StructureSymmetrizer


def append_interstitial(supercell_info: SupercellInfo,
                        unitcell_structure: Structure,
                        frac_coords: List[float]) -> SupercellInfo:
    """
    inv_trans_mat must be multiplied with coords from the right as the
    trans_mat is multiplied to the unitcell lattice vector from the left.
    see __mul__ of IStructure in pymatgen.
    x_u, x_s means the frac coordinates in unitcell and supercell,
    while a, b, c are the unitcell lattice vector.
    (a_u, b_u, c_u) . (a, b, c) = (a_s, b_s, c_s) . trans_mat . (a, b, c)
    (a_u, b_u, c_u) = (a_s, b_s, c_s) . trans_mat
    so, (a_s, b_s, c_s) = (a_u, b_u, c_u) . inv_trans_mat
    """
    if supercell_info.unitcell_structure and \
            supercell_info.unitcell_structure != unitcell_structure:
        raise NotPrimitiveError

    unitcell_structure.append(species=Element.H, coords=frac_coords)
    symmetrizer = StructureSymmetrizer(unitcell_structure)
    wyckoff_letter = (symmetrizer.spglib_sym_data["wyckoffs"][-1])
    site_symmetry = (symmetrizer.spglib_sym_data["site_symmetry_symbols"][-1])

    inv_matrix = inv(np.array(supercell_info.transformation_matrix))
    new_coords = np.dot(frac_coords, inv_matrix).tolist()
    supercell_info.interstitials.append(
        Interstitial(frac_coords=new_coords, wyckoff_letter=wyckoff_letter,
                     site_symmetry=site_symmetry))
    return supercell_info



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from pathlib import Path

from pydefect.cli.vasp.util_commands.make_refined_poscar import \
    make_refined_structure
from pydefect.defaults import defaults
from pydefect.input_maker.defect_entry import DefectEntry
from pymatgen.core import Structure


def test_make_refined_structure(tmpdir, mocker, simple_cubic):
    tmpdir.chdir()
    tmpdir.join("Va_O1_2").mkdir()
    simple_cubic.to(fmt="POSCAR", filename=defaults.contcar)
    tmpdir.join(defaults.outcar).write("")
    tmpdir.join(defaults.vasprun).write("")

    mock_defect_entry = mocker.Mock(spec=DefectEntry, autospec=True)

    def side_effect(key):
        if str(key) == "defect_entry.json":
            mock_defect_entry.anchor_atom_index = 0
            mock_defect_entry.structure = simple_cubic
            return mock_defect_entry
        else:
            raise ValueError

    mock_loadfn = mocker.patch("pydefect.cli.vasp.util_commands.make_refined_poscar.loadfn", side_effect=side_effect)
    mock = mocker.patch("pydefect.cli.vasp.util_commands.make_refined_poscar.symmetrize_defect_structure")
    mock.return_value = simple_cubic

    make_refined_structure()

    mock_loadfn.assert_any_call("defect_entry.json")

    assert Structure.from_file(Path("CONTCAR.sym_1")) == simple_cubic
    assert Structure.from_file(Path("POSCAR")) == simple_cubic



# -*- coding: utf-8 -*-

import os
import sys
from math import sqrt
from pathlib import Path

import pytest
from pydefect.corrections.efnv_correction import PotentialSite, \
    ExtendedFnvCorrection
from pydefect.defaults import defaults
from pydefect.input_maker.supercell_info import Site, SupercellInfo, \
    Interstitial
from pymatgen.core import Lattice, IStructure, Structure

sys.path.append(os.path.join(os.path.dirname(__file__), 'helpers'))
# Need the following the share the fixture


@pytest.fixture(scope="session")
def vasp_files():
    return Path(__file__).parent / "cli" / "vasp" / "vasp_files"


@pytest.fixture(scope="session")
def simple_cubic():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def simple_cubic_2x1x1():
    lattice = Lattice.orthorhombic(2.0, 1.0, 1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H", "H"], coords=coords)


@pytest.fixture(scope="session")
def simple_cubic_2x2x2():
    lattice = Lattice.cubic(2.0)
    coords = [[0.0, 0.0, 0.0],
              [0.5, 0.5, 0.0],
              [0.5, 0.0, 0.5],
              [0.0, 0.5, 0.5],
              [0.0, 0.0, 0.5],
              [0.0, 0.5, 0.0],
              [0.5, 0.0, 0.0],
              [0.5, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 8, coords=coords)


@pytest.fixture(scope="session")
def monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def complex_monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0],
              [0.1, 0.0, 0.0],
              [0.9, 0.0, 0.0],
              [0.2, 0.0, 0.0],
              [0.8, 0.0, 0.0]]
    return IStructure(lattice=lattice,
                      species=["H", "He", "He", "He", "He"],
                      coords=coords)


@pytest.fixture(scope="session")
def rhombohedral():
    lattice = Lattice.rhombohedral(a=1, alpha=45)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def elongated_tetragonal():
    lattice = Lattice.tetragonal(a=1, c=3 * sqrt(2))
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def elongated_tetragonal():
    lattice = Lattice.tetragonal(a=1, c=3 * sqrt(2))
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def a_centered_orthorhombic():
    lattice = Lattice([[1,  0, 0],
                       [0,  2, 3],
                       [0, -2, 3]])
    coords = [[0.5, 0.8, 0.8],
              [0.0, 0.3, 0.0],
              [0.0, 0.0, 0.3]]

    return IStructure(lattice=lattice, species=["H"] * 3, coords=coords)


@pytest.fixture(scope="session")
def c_centered_monoclinic():
    lattice = Lattice.monoclinic(3, 4, 5, 100)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]]
    return IStructure(lattice=lattice, species=["H", "H"], coords=coords)


@pytest.fixture(scope="session")
def fcc():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.0], [0.5, 0.0, 0.5], [0.0, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 4, coords=coords)


@pytest.fixture(scope="session")
def bcc():
    lattice = Lattice.cubic(1.0)
    coords = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5]]
    return IStructure(lattice=lattice, species=["H"] * 2, coords=coords)


@pytest.fixture(scope="session")
def tetra_close_to_cubic():
    lattice = Lattice.tetragonal(1.001 * 10 / sqrt(2), 10)
    coords = [[0.0, 0.0, 0.0]]
    return IStructure(lattice=lattice, species=["H"], coords=coords)


@pytest.fixture(scope="session")
def ortho_conventional():
    lattice = Lattice.orthorhombic(5, 6, 7)
    coords = [
        [0.0, 0.0, 0.0],
        [0.5, 0.5, 0.0],
        [0.5, 0.0, 0.5],
        [0.0, 0.5, 0.5],

        [0.0, 0.0, 0.5],
        [0.0, 0.5, 0.0],
        [0.5, 0.0, 0.0],
        [0.5, 0.5, 0.5],
    ]
    return IStructure(lattice=lattice, species=["H"] * 4 + ["He"] * 4,
                      coords=coords)


@pytest.fixture(scope="session")
def cubic_supercell():
    lattice = Lattice.cubic(10)
    coords = [
       [0.00, 0.00, 0.00],
       [0.00, 0.00, 0.50],
       [0.00, 0.50, 0.00],
       [0.00, 0.50, 0.50],
       [0.50, 0.00, 0.00],
       [0.50, 0.00, 0.50],
       [0.50, 0.50, 0.00],
       [0.50, 0.50, 0.50],
       [0.00, 0.25, 0.25],
       [0.00, 0.25, 0.75],
       [0.00, 0.75, 0.25],
       [0.00, 0.75, 0.75],
       [0.50, 0.25, 0.25],
       [0.50, 0.25, 0.75],
       [0.50, 0.75, 0.25],
       [0.50, 0.75, 0.75],
       [0.25, 0.00, 0.25],
       [0.25, 0.00, 0.75],
       [0.25, 0.50, 0.25],
       [0.25, 0.50, 0.75],
       [0.75, 0.00, 0.25],
       [0.75, 0.00, 0.75],
       [0.75, 0.50, 0.25],
       [0.75, 0.50, 0.75],
       [0.25, 0.25, 0.00],
       [0.25, 0.25, 0.50],
       [0.25, 0.75, 0.00],
       [0.25, 0.75, 0.50],
       [0.75, 0.25, 0.00],
       [0.75, 0.25, 0.50],
       [0.75, 0.75, 0.00],
       [0.75, 0.75, 0.50],
       [0.25, 0.00, 0.00],
       [0.25, 0.00, 0.50],
       [0.25, 0.50, 0.00],
       [0.25, 0.50, 0.50],
       [0.75, 0.00, 0.00],
       [0.75, 0.00, 0.50],
       [0.75, 0.50, 0.00],
       [0.75, 0.50, 0.50],
       [0.25, 0.25, 0.25],
       [0.25, 0.25, 0.75],
       [0.25, 0.75, 0.25],
       [0.25, 0.75, 0.75],
       [0.75, 0.25, 0.25],
       [0.75, 0.25, 0.75],
       [0.75, 0.75, 0.25],
       [0.75, 0.75, 0.75],
       [0.00, 0.00, 0.25],
       [0.00, 0.00, 0.75],
       [0.00, 0.50, 0.25],
       [0.00, 0.50, 0.75],
       [0.50, 0.00, 0.25],
       [0.50, 0.00, 0.75],
       [0.50, 0.50, 0.25],
       [0.50, 0.50, 0.75],
       [0.00, 0.25, 0.00],
       [0.00, 0.25, 0.50],
       [0.00, 0.75, 0.00],
       [0.00, 0.75, 0.50],
       [0.50, 0.25, 0.00],
       [0.50, 0.25, 0.50],
       [0.50, 0.75, 0.00],
       [0.50, 0.75, 0.50],
    ]
    return Structure(lattice=lattice, species=["H"] * 32 + ["He"] * 32,
                     coords=coords)


@pytest.fixture(scope="session")
def cubic_supercell_w_vacancy(cubic_supercell):
    result = cubic_supercell.copy()
    result.remove_sites([0])
    return result


@pytest.fixture(scope="session")
def cubic_supercell_w_vacancy_w_perturb(cubic_supercell_w_vacancy):
    result = cubic_supercell_w_vacancy.copy()
    result.translate_sites([0], [0.0, 0.0, 0.1])
    return result


@pytest.fixture
def supercell_info(mocker, ortho_conventional):
    mock = mocker.patch("pydefect.util.structure_tools.defaults")
    mock.dist_tol = defaults.dist_tol
    mock.cutoff_distance_factor = 1.7
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="mmm",
                        equivalent_atoms=[0, 1, 2, 3]),
             "He1": Site(element="He", wyckoff_letter="b", site_symmetry="mmm",
                         equivalent_atoms=[4, 5, 6, 7])}
    interstitial = Interstitial([0.25]*3, wyckoff_letter="x", site_symmetry="yy")
    return SupercellInfo(ortho_conventional,
                         "Fmmm",
                         [[1, 0, 0], [0, 1, 0], [0, 0, 1]], sites,
                         [interstitial])


@pytest.fixture
def cubic_supercell_info_wo_int(mocker, cubic_supercell):
    mock = mocker.patch("pydefect.util.structure_tools.defaults")
    mock.dist_tol = defaults.dist_tol
    mock.cutoff_distance_factor = 1.7
    sites = {"H1": Site(element="H", wyckoff_letter="a", site_symmetry="m-3m",
                        equivalent_atoms=[0, 1, 2, 3]),
             "He1": Site(element="He", wyckoff_letter="b", site_symmetry="m-3m",
                         equivalent_atoms=[4, 5, 6, 7])}
    unitcell = Structure(Lattice.rhombohedral(7.071068, 60),
                         species=["H", "He"],
                         coords=[[0.0]*3, [0.5]*3])
    return SupercellInfo(cubic_supercell,
                         "Fm-3m", [[-2, 2, 2], [2, -2, 2], [2, 2, -2]], sites,
                         unitcell_structure=unitcell)


@pytest.fixture
def efnv_correction():
    s1 = PotentialSite(specie="H", distance=1.999, potential=1.0, pc_potential=None)
    s2 = PotentialSite(specie="He", distance=2.0001, potential=1.5, pc_potential=0.2)
    s3 = PotentialSite(specie="He", distance=3.0, potential=2.0, pc_potential=0.3)

    return ExtendedFnvCorrection(charge=10,
                                 point_charge_correction=1.0,
                                 defect_region_radius=2.0,
                                 sites=[s1, s2, s3],
                                 defect_coords=(0.0, 0.0, 0.0),
                                 additional_correction=100)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

import fire
from pymatgen.core import Element, Structure
from pymatgen.analysis.defects.utils import ChargeDensityAnalyzer
from pymatgen.io.vasp import Chgcar, VolumetricData
from vise.util.structure_symmetrizer import StructureSymmetrizer


def interstitials_from_volumetric_data(
        volumetric_data: VolumetricData,
        find_min: bool = True,
        threshold_frac: float = None,
        threshold_abs: float = None,
        min_dist: float = 0.5,
        tol: float = 0.5,
        radius: float = 0.4) -> None:

    cda = ChargeDensityAnalyzer(chgcar=volumetric_data)
    cda.get_local_extrema(threshold_frac=threshold_frac,
                          threshold_abs=threshold_abs,
                          find_min=find_min)
    cda.sort_sites_by_integrated_chg(r=radius)
    # Remove sites near host atoms.
    cda.remove_collisions(min_dist)
    # Cluster interstitials that are too close together using a tol.
    try:
        cda.cluster_nodes(tol=tol)
    except ValueError:
        # error raised when only a single local extreme point.
        pass
    print(cda.extrema_df)

    structure: Structure = volumetric_data.structure.copy()
    start_index = len(structure)
    end_index = len(structure) + len(cda.extrema_coords)
    interstitial_indices = [i for i in range(start_index, end_index)]
    for coords in cda.extrema_coords:
        structure.append(Element.H, coords)
    symmetrizer = StructureSymmetrizer(structure)
    equiv_atoms = symmetrizer.spglib_sym_data["equivalent_atoms"]

    print(f"Host symmetry {symmetrizer.spglib_sym_data['international']}")
    print("++ Inequivalent indices and site symmetries ++")
    orig_num_atoms = len(volumetric_data.structure)

    structure_for_visualize: Structure = volumetric_data.structure.copy()

    for i, ii in enumerate(interstitial_indices):
        if ii == equiv_atoms[ii]:
            idx = orig_num_atoms + i
            coords = structure[idx].frac_coords

            def element():
                for z in range(1, 100):
                    e = Element.from_Z(z)
                    if e not in structure_for_visualize.composition:
                        yield e

            structure_for_visualize.append(next(element()), coords)

            idx_coords = \
                f"{i:>3} {coords[0]:8.4f} {coords[1]:8.4f} {coords[2]:8.4f}"
            print(idx_coords,
                  symmetrizer.spglib_sym_data["site_symmetry_symbols"][ii])

        structure_for_visualize.to(filename="POSCAR-w-interstitials")


def interstitials_from_charge_density(aeccar0, aeccar2, **kwargs):
    aeccar = Chgcar.from_file(aeccar0) + Chgcar.from_file(aeccar2)
    interstitials_from_volumetric_data(aeccar, **kwargs)


if __name__ == '__main__':
    fire.Fire(interstitials_from_charge_density)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.
from typing import Optional

import numpy as np
from numpy.linalg import inv
from pydefect.analyzer.defect_structure_info import logger
from pymatgen.core import Structure
from vise.util.structure_symmetrizer import StructureSymmetrizer


def symmetrize_defect_structure(structure_symmetrizer: StructureSymmetrizer,
                                anchor_atom_idx: Optional[int] = None,
                                anchor_atom_coord: Optional[np.ndarray] = None
                                ) -> Structure:
    result = structure_symmetrizer.structure.copy()

    logger.info(f"The symmetry is {structure_symmetrizer.point_group}")

    origin_shift = structure_symmetrizer.spglib_sym_data["origin_shift"]
    inv_trans_mat = inv(structure_symmetrizer.spglib_sym_data["transformation_matrix"])
    coords = structure_symmetrizer.spglib_sym_data["std_positions"]
    new_coords = []
    for i in range(len(result)):
        new_coords.append(np.dot(inv_trans_mat, (coords[i] - origin_shift)))

    if anchor_atom_idx:
        offset = new_coords[anchor_atom_idx] - anchor_atom_coord
    else:
        offset = np.array([0.0]*3)
    new_coords = np.array(new_coords) - offset
    for i, coords in zip(result, new_coords):
        i.frac_coords = coords % 1
    return result


# -*- coding: utf-8 -*-
#  Copyright (c) 2020 Kumagai group.

from pydefect.cli.vasp.util_commands.recommend_interstitials import \
    interstitials_from_volumetric_data
from pymatgen.io.vasp import Chgcar


def test_interstitials_from_charge_density(vasp_files):
    aeccar0 = Chgcar.from_file(str(vasp_files / "NaMgF3_AECCAR0"))
    aeccar2 = Chgcar.from_file(str(vasp_files / "NaMgF3_AECCAR2"))
    aeccar = aeccar0 + aeccar2
    aeccar.write_file("CHGCAR")
    interstitials_from_volumetric_data(aeccar)

#
# def test_interstitials_from_potential():
#     locpot = Locpot.from_file(str(Path(__file__).parent / "vasp_files" / "NaMgF3_LOCPOT"))
#     interstitials_from_charge_density(locpot)

